/*************************************************************************
 * This file is part of Crosroads 1.79, a load balancer and fail over
 * utility for TCP. Copyright (c) Karel Kubat, distributed under GPL.
 * Visit http://crossroads.e-tunity.com for information.
 *************************************************************************/

/* Actually used commandline size. */
#ifndef SPT_BUFSIZE
#define SPT_BUFSIZE DEFAULT_SPT_BUFSIZE
#endif

/* Includes. */
#include <ctype.h>
#include <errno.h>
#include <fcntl.h>
#include <netdb.h>
#if HAVE_MALLOC_H == 1
#include <malloc.h>
#endif
#include <math.h>
#include <pwd.h>
#include <signal.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#if HAVE_STDINT_H == 1
#include <stdint.h>
#endif
#include <string.h>
#include <syslog.h>
#include <time.h>
#include <unistd.h>
#include <arpa/inet.h>
#include <netinet/in.h>
#include <sys/file.h>
#include <sys/ipc.h>
#include <sys/resource.h>
#include <sys/sem.h>
#include <sys/shm.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/times.h>
#include <sys/types.h>

/* flock() macros, incase they aren't present on your system. */
#ifndef LOCK_SH
#define LOCK_SH 1
#endif
#ifndef LOCK_EX
#define LOCK_EX 2
#endif
#ifndef LOCK_NB
#define LOCK_NB 4
#endif
#ifndef LOCK_UN
#define LOCK_UN 8
#endif

/* inet_addr() macros, incase they aren't present on your system. */
#ifndef INADDR_NONE
#define INADDR_NONE ((unsigned long) -1)
#endif

/* Types. */

/* NOTE: Adapt optionsread.c and data.c and parser.y when changing!
 * -------------------------------------------------------------- */
typedef struct {			/* Operating options in daemon */
    int log_activity;			/* -a: log TCP timings? */
    unsigned tcp_bufsz;			/* -B: TCP buffer size */
    int dns_cachettl;			/* -d: DNS cache timeout */
    int log_facility;			/* -l: logging facility */    
    int shmperm;			/* -m: SHM permissions */
    int sloppyportbind;			/* -s: sloppy port bindings? */
    int leave_proctitle;		/* -p: leave process title? */
    int log_http_activity;		/* -h: log HTTP activity? */
} Options;

typedef struct {			/* Symbol table for #define */
    char *symbol;
    char *redef;
} Symtab;

typedef struct {			/* Bison/Flex state given files */
    char *name;
    FILE *file;
    int lineno;
    char *buf;
} ParserState;

typedef enum {				/* Config parsing related */
    cf_portspec,
    cf_backlogspec,
    cf_verbosityspec,
    cf_backendspec,
    cf_serverspec,
    cf_dispatchspec,
    cf_revivespec,
    cf_checkspec,
    cf_shmkeyspec,
    cf_weightspec,
    cf_decayspec,
    cf_onstartspec,
    cf_onfailspec,
    cf_onendspec,
    cf_connectiontimeoutspec,
    cf_maxconnectionsspec,
    cf_typespec,
    cf_inspectionspec,
    cf_dumpspec,
    cf_thruspec,
    cf_httptimingspec,
    cf_bindspec,
    cf_stickycookiespec,
    cf_setclientheaderspec,
    cf_addclientheaderspec,
    cf_appendclientheaderspec,
    cf_setserverheaderspec,
    cf_addserverheaderspec,
    cf_appendserverheaderspec,
    cf_allowfromspec,
    cf_denyfromspec,
    cf_allowfilespec,
    cf_denyfilespec,
    cf_useraccountspec,
    cf_retriesspec,
    cf_statespec,
} Conftype;

typedef union {				/* Integer of string value */
    int ival;				/* passed around by the parser */
    char *sval;
} IntOrString;

typedef struct {
    Conftype cf;
    IntOrString v;
} Confset;

typedef struct {
    int n;
    Confset *set;
} Confsets;

typedef enum {				/* Dispatching types */
    ds_roundrobin,
    ds_random,
    ds_bysize,
    ds_byduration,
    ds_byorder,
    ds_byconnections,
    ds_byclientip,
    ds_externalhandler
} Dispatchtype;

typedef enum {				/* Service types */
    type_any,
    type_http
} Servicetype;

typedef enum {				/* Backend availability */    
    st_available,			/* .. all OK */
    st_unavailable,			/* .. temporarily unavailable */
    st_down,				/* .. permanently unavailable */
    st_waking,				/* .. scheduled for wake up */
    st_intermediate,			/* .. bogus, for mark_activity() */
    st_unknown 				/* .. no idea (terminator) */
} Backendavail;
#define NR_OF_STATES 6			/* Make sure this matches the above */
					/* possible states! */

typedef struct {			/* Backend description */
    char *name;				/* .. back end identifier */
    int initial_port;			/* .. TCP port */
    char *initial_server;		/* .. server name */
    int verbosity;			/* .. debugging on/off */
    int weight;				/* .. weight in backend selection */
    int decay;				/* .. decay in % */
    char *onstart;			/* .. system() on success */
    char *onfail;			/* .. or on failure */
    char *onend;			/* .. or on ending */
    char *dumpfile;			/* .. traffic dump file */
    char *thruputfile;			/* .. traffic throughput file */
    char *timinglog;			/* .. HTTP timings log */
    unsigned maxconnections;		/* .. max # of allowed connections */
    char *stickycookie;			/* .. cookie selector */
    int retries;			/* .. # retries before unavailable */
    char **addclientheader;		/* .. client hdrs to ADD */
    int naddclientheader;		/* .. table size */
    char **setclientheader;		/* .. client hdrs to SET */
    int nsetclientheader;		/* .. table size */
    char **appendclientheader;		/* .. client hdrs to APPEND */
    int nappendclientheader;		/* .. table size */
    char **addserverheader;		/* .. server hdrs to ADD */
    int naddserverheader;		/* .. table size */
    char **setserverheader;		/* .. server hdrs to SET */
    int nsetserverheader;		/* .. table size */
    char **appendserverheader;		/* .. server hdrs to APPEND */
    int nappendserverheader;		/* .. table size */
    Backendavail initial_avail;		/* .. initial back end state */
} Backend;

typedef struct {			/* Filtering information: */
    unsigned ip;			/* .. IP address (or part of it) */
    unsigned mask;			/* .. netmask */
} IpFilter;

typedef enum {				/* HTTP header inspection mode */
    ins_deep,				/* .. inspect all headers */
    ins_shallow,			/* .. inspect only first */
} HttpInspection;

/* The configuration in internal representation. When changing, remember
 * to change config_write() and config_read() and config_msg()! */
typedef struct {			/* Service description */
    /* Basic stuff */
    char *name;				/* .. service name */
    char *bind;				/* .. address to bind to */
    unsigned port;			/* .. listening port */
    unsigned verbosity;			/* .. message generation */
    Dispatchtype dispatchtype;		/* .. backend selection method */
    unsigned dispatchover;		/* .. selection over # connections */
    char *dispatchext;			/* .. external handler */ 
    unsigned rev_interval;		/* .. dead backend revival interval */
    unsigned check_interval;		/* .. backend recheck interval */
    char *check_cmd;			/* .. rechecking command */
    unsigned backlog;			/* .. # pending TCP connections */
    unsigned shmkey;			/* .. key for SysV IPC */
    unsigned connectiontimeout;		/* .. # secs for timeout handling */
    unsigned maxconnections;		/* .. max # of allowed connections */
    Servicetype type;			/* .. type of the service */
    HttpInspection inspection;		/* .. header inspection mode */
    char *allowfile;			/* .. file with allowed IP filters */
    char *denyfile;			/* .. and denied */
    int uid;				/* .. UID to assume for system() */
    int gid;				/* .. GID to assume for system() */
    /* Deny chain */
    int ndenychain;			/* .. size of chain */
    IpFilter *denychain;		/* .. chain of denied IP's */
    /* Allow chain */
    int nallowchain;			/* .. size of chain */
    IpFilter *allowchain;		/* .. chain of allowed IP's */
    /* Back ends */
    int nbackend;			/* .. size of backend array */
    Backend *backend;			/* .. the back ends */
} Service;

typedef struct {			/* Backend state */
    Backendavail actual_avail;		/* .. availability */
    unsigned long totuses;		/* .. times hit */
    unsigned long failures;		/* .. times failed */
    double nbytes;			/* .. transferred bytes */
    unsigned long sessions;		/* .. # of http sessions */
    unsigned long avg_nbytes;		/* .. averaged over # connections */
    double nsec;			/* .. connection durations */
    double avg_nsec;			/* .. averaged over # connections */
    unsigned nclients;			/* .. active clients */
    unsigned fail;			/* .. failures so far */
    unsigned laststamp;			/* .. last usage */
    int actual_port;			/* .. currently used port */
    char actual_server[FQDN_LENGTH];	/* .. currently used server */
    double last_con_sec;		/* .. last connect duration */
} Backendstate;

typedef struct {			/* Service reporting (shmem) */
    int pid;				/* .. PID of service daemon */
    int rev_pid;			/* .. PID of revival handler */
    unsigned nclients;			/* .. active clients */
    int last_backend;			/* .. last used back end */
    Backendstate			/* .. states of the back ends */
	backendstate[1];		/* .. actual shmem array will be >1 */
} Servicereport;

typedef enum {				/* Stage of the program: */
    stage_main,				/* .. commandline */
    stage_waiting,			/* .. waiting for connections */
    stage_serving,			/* .. servicing a connection */
    stage_retrying			/* .. waking up a backend */
} Programstage;				/* Update stagetostring.c when */
					/* modifying! */

typedef struct {			/* An HTTP Message */
    char **header;			/* Headers table, excluding first */
    int nheader;			/* Headers table size */
} HttpHeader;

typedef enum {				/* Time representation format */
    tm_localtime,			/* .. local */
    tm_gmtime				/* .. UTC */
} TmType;

typedef enum {				/* TCP copying direction */
    dir_client_to_server,
    dir_server_to_client
} CopyDirection;

typedef enum {				/* HTTP version */
    con_unknown,
    con_keepalive,
    con_close
} HttpConnectionType;

typedef struct {			/* Numeric state to string map */
    Backendavail av;
    char *nm;
} StateStringMap;

typedef struct {			/* DNS cache related */
    char *hostname;			/* resolved name */
    struct in_addr hostaddr;		/* resolved host info entry */
    time_t stamp;			/* date of cache entry */
} DnsCache;

/* Globals. */
#ifndef EXTERN
#define EXTERN extern
#endif

EXTERN Service *activeservice;		/* target service of a daemon */
EXTERN unsigned char *clbuf;		/* client socket input buffer */
EXTERN int clbufpos, clbufmax;		/* .. position & bytes */
EXTERN char *client_ip;			/* connected client */
EXTERN unsigned client_ip_nr;		/* .. and in reversed octets form */
EXTERN char *config_file;		/* config to parse */
EXTERN int current_backend;		/* of a given service */
EXTERN int daemonized;			/* are we forked off yet */
EXTERN DnsCache dnscache[];		/* DNS cache */
EXTERN int dns_cachettl;		/* TTL of DNS cache entries */
EXTERN int flag_verbose;		/* flag: verbosity */
EXTERN int gid_org, gid_set;		/* original gid, flag: changed? */
EXTERN int interrupted;			/* got a signal? */
EXTERN char *laststring;		/* semantic lexer value */
EXTERN int listen_sock;			/* servicer listening socket */
EXTERN int logstarted;			/* was syslog() called yet? */
EXTERN int nservice;			/* # service descriptions */
EXTERN Options opt;			/* processing options */
EXTERN int org_argc;			/* original argc */
EXTERN char **org_argv;			/* and original argv */
EXTERN int parser_debug;		/* lexer/parser debugging */
EXTERN ParserState *parserstate;	/* parser state */
EXTERN int nparserstate;		/* size of array */
EXTERN char *program_name;		/* fully expanded argv[0] */
EXTERN Programstage program_stage;	/* stage of the program */
EXTERN int relevant_sigs[];		/* relevant signals */
EXTERN int reload_allow_deny;		/* flag to reload spec files */
EXTERN int semid;			/* semaphore ID */
EXTERN Service *service;		/* service descriptions */
EXTERN Servicereport *servicereport;	/* reporter in shared mem */
EXTERN unsigned char *srbuf;		/* server socket input buffer */
EXTERN int  srbufpos, srbufmax;		/* .. position & bytes */
EXTERN StateStringMap			/* state/string map */
    statestringmap[];			/* .. and vv */
EXTERN Symtab *symtab;			/* symbol table of #defines */
EXTERN int nsymtab;			/* table size */
EXTERN int tabular_status;		/* -t: tabular status overview */
EXTERN int uid_org, uid_set;		/* original uid, flag: changed? */
EXTERN int xml_status;			/* -x: XML status overview */
EXTERN char *xml_xslt;			/* -X: XSLT refrence for XML output */
EXTERN FILE *yyin;			/* config file handle */
EXTERN int yylineno;			/* input line number */
EXTERN char *yyerrmsg;			/* parsing error msg */
EXTERN char *yytext;			/* lexical buffer */

/* Functions. To suppress the __attribute__ macro (if you're on a
 * non-gcc compiler), uncomment the following. */
// #undef __attribute__
// #define __attribute(x)
extern void alloc_reporter (Service *active, int first);
extern char *ansistamp (TmType t, int offset);
extern int backend_available (int target);
extern int backend_count(void);
extern int backend_connect (void);
extern void check_service (Service s);
extern void choose_backend (void);
extern int cmd_restart (int ac, char **av);
extern int cmd_services (int ac, char **av);
extern int cmd_status (int ac, char **av);
extern int cmd_start (int ac, char **av);
extern int cmd_stop (int ac, char **av);
extern int cmd_tell (int ac, char **av);
extern int cmd_configtest (int ac, char **av);
extern void config_msg (void);
extern void config_read (int fd);
extern void config_write (int fd);
extern void copysockets (int clientsock, int serversock);
extern void decr_client_count (void);
extern void dealloc_reporter (Service *s);
extern struct in_addr *dns (char const *hostname);
extern void error (char const *fmt, ...)
    __attribute__ ((format (printf, 1, 2)));
extern void expand_program_name (char const *av0);
extern char *gmtstamp (int offset);
extern unsigned hashpjw (char const *s);
extern void http_copy (HttpHeader *h, int src_sock, int dst_sock,
		       CopyDirection dir);
extern void http_error (int clientsock);
extern void http_header_addheader (HttpHeader *m, char const *hdr);
extern void http_header_appendheader (HttpHeader *m, char const *hdr);
extern HttpConnectionType http_header_connectiontype (HttpHeader *h);
extern void http_header_free (HttpHeader *msg);
extern int http_header_hascookie (HttpHeader *m, char const *cookie);
extern double http_header_httpver (HttpHeader *h);
extern void http_header_removeheader (HttpHeader *h, char const *hdr);
extern unsigned char const *http_header_val (HttpHeader *h, char const *hdr);
extern HttpHeader *http_header_new (void);
extern void http_header_read (HttpHeader *m, int srcsock, CopyDirection dir);
extern void http_header_setheader (HttpHeader *m, char const *hdr);
extern void http_header_write (HttpHeader *m, int sock, CopyDirection dir);
extern void http_log_activity (double start, double end, char const *fmt, ...);
extern void http_serve (int clientsock);
extern int http_serversocket (HttpHeader *m, int *is_continuation);
extern int http_write (int sock, CopyDirection dir, unsigned char const *buf,
		       int buflen);
extern void incr_client_count (void);
extern int init_sockaddr (struct sockaddr_in *name,
			   const char *hostname, int port);
extern void interrupt (int sig);
extern int ipf_add_allow (Service *s, char const *val);
extern int ipf_add_deny (Service *s, char const *val);
extern int ipf_allowed (void);
extern int ipf_denied (void);
extern int ipf_loadfile (char const *fname, IpFilter **chain, int *nchain);
extern int ipf_match (IpFilter f);
extern int ipf_parse (char const *val, IpFilter *res);
extern int is_hex_digit (char ch);
extern int is_space (char ch);
extern void limits_msg (void);
extern void lock_reporter (void);
extern void log_activity_any (char const *action);
extern void log_activity_start (void);
extern void log_activity_end (void);
extern void log_activity_continuation (void);
extern int make_socket (int port, char const *ipaddr);
extern void mark_activity (double nbytes, double nsec, Backendavail newstate);
extern void msg (char const *fmt, ...)
    __attribute__ ((format (printf, 1, 2)));
extern void msgdumpbuf (unsigned char const *buf, int buflen);
extern int net_copy (int clientsock, int serversock, int maxbytes,
			  unsigned char *buf);
extern int net_read (int sock, int maxbytes, unsigned char *buf,
		     int is_client);
extern void net_allocbufs (void);
extern unsigned char *net_buffer (CopyDirection dir, unsigned *sz);
extern unsigned char *net_bufread (int sock, int maxbytes,
				   int *nreadp, int is_client);
extern int net_write (int sock, unsigned char const *buf, int len,
		      int is_client);
extern void options_read (int fd);
extern void options_write (int fd);
extern int parser_close();
extern char const *parser_filename(void);
extern void parser_open (char const *name);
extern int parser_input (char *buf, int max);
extern void parser_run(void);
extern void parser_skipchar(void);
extern void parser_skipline(void);
extern void runservice (void);
extern void set_program_title (char const *fmt, ...)
    __attribute__ ((format (printf, 1, 2)));
extern void show_config (void);
extern char *stage_to_string (Programstage stage);
extern char *state_to_string (Backendavail avail);
extern Backendavail string_to_state (char const *str);
extern char *str_expand_format (char const *h);
extern char *str_printf (char const *fmt, ...)
    __attribute__ ((format (printf, 1, 2)));
extern char *str_vprintf (char const *fmt, va_list args);
extern void symtab_end (void);
extern char const *symtab_lookup (char const *txt);
extern void symtab_set (char const *symbol);
extern void symtab_start (char const *symbol);
extern int sysrun (char const *cmd);
extern void tcpserve (int sock);
extern double timeofday(void);
extern void thruputlog (unsigned char const *buf, int len, CopyDirection dir);
extern void trafficlog (unsigned char const *buf, int len, CopyDirection dir);
extern void unlock_reporter (void);
extern void uid_assume (void);
extern void uid_restore (void);
extern void usage (void);
extern void wakeup_handler (void);
extern void warning (char const *fmt, ...)
    __attribute__ ((format (printf, 1, 2)));
extern void writelog (int prio, char const *fmt, ...)
    __attribute__ ((format (printf, 2, 3)));
extern int yyparse (void);
extern void *xcalloc (unsigned sz);
extern void *xmalloc (unsigned sz);
extern void *xrealloc (void *what, unsigned sz);
extern char *xstrcat (char *what, char const *rest);
extern char *xstrcatch (char *what, char ch);
extern char *xstrdup (char const *what);

/* strlcat() if it's missing on this system */
#if HAVE_STRLCAT == 0
extern size_t strlcat (char *dst, char const *src, size_t size);
#endif

/* vsyslog() if it's missing on this system */
#if HAVE_VSYSLOG == 0
extern void vsyslog (int fac, char const *fmt, va_list args);
#endif

/* strcasestr() if it's missing on this system */
#if HAVE_STRCASESTR == 0
extern char *strcasestr (char const *big, char const *little);
#endif

/* setproctitle() if it's missing on this system */
#if HAVE_SETPROCTITLE == 0
extern void setproctitle (char const *fmt, ...);
#endif

/* strupr() if it's missing on this system */
#if HAVE_STRUPR == 0
extern char *strupr (char *buf);
#endif

/* strnstr() if it's missing on this system */
#if HAVE_STRNSTR == 0
extern char *strnstr (char const *buf, char const *little, size_t len);
#endif

/* __progname_full on systems that have it */
#if HAVE_PROGNAME_FULL == 1
extern char *__progname_full;
#endif

