/*************************************************************************
 * This file is part of Crosroads 1.79, a load balancer and fail over
 * utility for TCP. Copyright (c) Karel Kubat, distributed under GPL.
 * Visit http://crossroads.e-tunity.com for information.
 *************************************************************************/

#include "../crossroads.h"

static int timed_out;

static void alarmhandler (int sig) {
    timed_out++;
}

int backend_connect () {
    int sock;
    struct sockaddr_in servername;
    int val = 1;
    struct timeval con_start, con_end;

    /* Create the socket. */
    if ( (sock = socket (PF_INET, SOCK_STREAM, 0)) < 0 ) {
	if (program_stage != stage_retrying) {
	    error ("failed to create socket for backend communication: %s",
		   strerror(errno));
	} else {
	    warning ("failed to create socket for backend communication: %s",
		     activeservice->name, strerror(errno));
	    return (-1);
	}
    }

#ifdef SO_NOSIGPIPE
    /* Make sure we don't get SIGPIPE's on it. Not supported on all
     * platforms. */
    if (setsockopt (sock, SOL_SOCKET, SO_NOSIGPIPE, &val, sizeof(val))) {
	warning ("cannot set backend socket options (nosigpipe): %s",
		 activeservice->name, strerror(errno));
	close (sock);
	return (-2);
    }
#endif

    /* Assign to back end server address. */
    if (init_sockaddr (
	    &servername,
	    servicereport->backendstate[current_backend].actual_server,
	    servicereport->backendstate[current_backend].actual_port)) {
	/* The hostname is unusable.. */
	warning ("unknown host %s",
		 servicereport->backendstate[current_backend].actual_server);
	close (sock);
	mark_activity (0, 0, st_unavailable);
	return (-3);
    }
    if (program_stage != stage_retrying)
	msg ("Network socket to %s:%d created",
	     servicereport->backendstate[current_backend].actual_server,
	     servicereport->backendstate[current_backend].actual_port);

    /* If retrying, delay a bit, unless we're in the wakeup handler. */
    if (program_stage != stage_retrying &&
	servicereport->backendstate[current_backend].fail > 0) {
	msg ("Seen %d retries, waiting a bit",
	     servicereport->backendstate[current_backend].fail);
	sleep (RETRY_WAIT);
    }

    /* Set up the connect interruption */
    signal (SIGALRM, alarmhandler);
    timed_out = 0;
    alarm (CONNECT_TIMEOUT);

    /* Try connecting. Remember the time we try this. */
    gettimeofday (&con_start, 0);
    if (connect (sock, (struct sockaddr *) &servername,
		 sizeof(servername)) < 0) {
	/* This backend is unusable.
	 * Either connect() has failed (connection refused) or it got
	 * interrupted by the SIGALRM, or another signal went off.
	 */
	close (sock);
	alarm(0);
	signal (SIGALRM, SIG_DFL);

	if (program_stage != stage_retrying) {
	    if (timed_out)
		warning (
		    "server %s:%d not usable due to timeout",
		    servicereport->backendstate[current_backend].actual_server,
		    servicereport->backendstate[current_backend].actual_port);
	    else
		warning (
		    "server %s:%d: cannot connect: %s",
		    servicereport->backendstate[current_backend].actual_server,
		    servicereport->backendstate[current_backend].actual_port,
		    strerror(errno));
	}
	mark_activity (0, 0, st_unavailable);
	return (-3);
    }
    alarm(0);
    signal (SIGALRM, SIG_DFL);

    /* Log the last connect time. Store the duration of this connect. */
    gettimeofday (&con_end, 0);
    servicereport->backendstate[current_backend].last_con_sec =
	( ((double)con_end.tv_sec * 1000000 + con_end.tv_usec) -
	  ((double)con_start.tv_sec * 1000000 + con_start.tv_usec) )
	/ 1000000;
    servicereport->backendstate[current_backend].laststamp = time(0);

    return (sock);
}
