/*************************************************************************
 * This file is part of Crosroads 1.79, a load balancer and fail over
 * utility for TCP. Copyright (c) Karel Kubat, distributed under GPL.
 * Visit http://crossroads.e-tunity.com for information.
 *************************************************************************/

#include "../crossroads.h"

void choose_backend () {
    int *backends = 0, nbackends = 0, h, i, j, k,
	target_set = 0, flat_weights = 1, *weights = 0, *sel_weights,
	tot_weights, lo_val, hi_val, done;
    double *values = 0, nbest;
    unsigned nclients;
    char *exthandler;
    FILE *f;
    char buf[80];
#   if  HAVE_SRANDDEV == 0
    struct timeval tv;
#   endif

    /* Check that we're allowed to accept at all. */
    if (activeservice->maxconnections &&
	servicereport->nclients >= activeservice->maxconnections) {
	warning ("max clients %d exceeded", activeservice->maxconnections);
	current_backend = -1;
	return;
    }

    /* Incase of dispatching by client IP, check that the back end
     * is up. If it is, take that target. */
    if (activeservice->dispatchtype == ds_byclientip) {
	h = hashpjw (client_ip) % activeservice->nbackend;
	msg ("Client IP %s hashed to back end index %d", client_ip, h);
	if (servicereport->backendstate[h].actual_avail == st_available &&
	    (activeservice->backend[h].maxconnections == 0 ||
	     activeservice->backend[h].maxconnections >
	     servicereport->backendstate[h].nclients)) {
	    msg ("Hashed back end %d is acceptable", h);
	    current_backend = h;
	    return;
	} else
	    msg ("Hashed back end %d is not available, redispatching", h);
    }

    /* Populate the array of selectable backends. */
    for (i = 0; i < activeservice->nbackend; i++) {
	/* Backend i is a candidate if:
	 * (a) it is available, AND
	 * (b) there's no max of connections, or the actual nr of connections
	 *     is below max
	 */
	if (servicereport->backendstate[i].actual_avail == st_available &&
	    (activeservice->backend[i].maxconnections == 0 ||
	     activeservice->backend[i].maxconnections >
	     servicereport->backendstate[i].nclients)) {
	    msg ("candidate back end: %d (max clients %d, active %d, "
		 "state %s)",
		 i, activeservice->backend[i].maxconnections,
		 servicereport->backendstate[i].nclients,
		 state_to_string(servicereport->backendstate[i].actual_avail));
	    backends = xrealloc (backends, (nbackends + 1) * sizeof(int) );
	    backends[nbackends++] = i;
	}
    }
    
    /* When we don't have any active back ends, then we sleep a bit.
     * Abort choosing a back end, unless there is an external dispatcher -
     * it may overrule our assumptions. */
    if (!nbackends) {
	sleep (RETRY_WAIT);
	if (activeservice->dispatchtype != ds_externalhandler) {
	    current_backend = -1;
	    warning ("No active backends to select!");
	    return;
	}
    }

    /* Only one backend? Then it's always the first one, whatever
     * you do. Exception is when we have an external dispatcher. In that
     * case we want to call it, even when there's just one back end.
     * Thanks Nicolas Prochazka for suggesting this. */
    if (nbackends == 1 && activeservice->dispatchtype != ds_externalhandler) {
	current_backend = backends[0];
	servicereport->last_backend = current_backend;
	msg ("only 1 backend to select (%d)", current_backend);
	free (backends);
	return;
    }

    /* More than 1 backends available, or external dispatching.
       Make a wise choice. */
    switch (activeservice->dispatchtype) {

    case ds_roundrobin:
	/* Find the currently used backend, go one forward. */
	msg ("last roundrobin back end was %d", servicereport->last_backend);
	for (i = 0; i < nbackends; i++)
	    if (backends[i] == servicereport->last_backend) {
		i++;
		i %= nbackends;
		current_backend = backends[i];
		servicereport->last_backend = current_backend;
		msg ("next roundrobin back end is %d", current_backend);
		free (backends);
		return;
	    }
	/* None found.. try the first one (run to momma). */
	current_backend = backends[0];
	servicereport->last_backend = current_backend;
	msg ("chosen backend (roundrobin, without historical info): %d",
	     current_backend);
	free (backends);
	return;
	
    case ds_random:
	/* Re-randomize. */
#	if HAVE_SRANDDEV == 1
	sranddev();
	msg ("randomizer seeded with randdev");
#	else	
	gettimeofday (&tv, 0);
	srand ((unsigned) tv.tv_usec);
	msg ("randomizer seeded with %u", (unsigned) tv.tv_usec);
#	endif
	
	/* First of all let's see if all the weights are the same. */
	values  = xmalloc (nbackends * sizeof(double));
	weights = xmalloc (nbackends * sizeof(int));
	tot_weights = 0;
	for (i = 0; i < nbackends; i++) {
	    values[i] = activeservice->backend[backends[i]].weight;
	    if (activeservice->backend[backends[i]].weight != 1)
		flat_weights = 0;
	}

	if (! flat_weights) {
	    /* Non-flat weight random balancing. Determine the chance
	     * to hit a given back end. First we populate the weights array.
	     * Eg imagine we have 4 back ends with weights 1,2,3,4. Then
	     * ultimately we need to get a selection array of indices:
	     * 0000111223, meaning that the chance to hit index 0 is 4/10
	     * and so on */

	    /* First of all we reverse the values vector, to get 4,3,2,1 */
	    done = 0;
	    while (!done) {
		done = 1;
		lo_val = 9999999;
		hi_val = 0;
		/* Find highest/lowest weights in values vector. */
		for (i = 0; i < nbackends; i++) {
		    if (!values[i])
			continue;
		    done = 0;
		    if (lo_val >= values[i])
			lo_val = values[i];
		    if (hi_val <= values[i])
			hi_val = values[i];
		}
		
		if (done)
		    break;

		/* Stick the values into the duplicate 'weights' array,
		   but the other way 'round. */
		for (i = 0; i < nbackends; i++)
		    if (values[i] == lo_val)
			weights[i] = hi_val;
		    else if (values[i] == hi_val)			
			weights[i] = lo_val;

		for (i = 0; i < nbackends; i++)
		    if (values[i] == lo_val) {
			tot_weights += hi_val;
			values[i] = 0;
		    } else if (values[i] == hi_val) {
			tot_weights += lo_val;
			values[i] = 0;
		    }
	    }
	    
	    /* The 'weights' vector is now the reverse of 'values'.
	     * Create a selection array. */
	    sel_weights = xmalloc (tot_weights * sizeof(int));
	    k = 0;
	    for (i = 0; i < nbackends; i++)
		for (j = 0; j < weights[i]; j++) {
		    sel_weights[k] = i;
		    k++;
		}

	    /* Debugging: 
	     * for (k = 0; k < tot_weights; k++)
	     *	 msg ("random: sel_weight[%d] = %d", k, sel_weights[k]);
	     */
	    
	    /* Select the next random back end. Algorithm is:
	     * i = rand() % tot_weights;
	     * j = sel_weights[i];
	     * k = backends[j];
	     * current_backend = k;
	     * msg ("i=%d j=%d, k=%d", i, j, k);
	     */
	    i = rand() % tot_weights;
	    current_backend = backends[sel_weights[i]];
	    servicereport->last_backend = current_backend;
	    free (sel_weights);
	    msg ("chosen backend (weighted random): %d at index %d",
		 current_backend, i);
	    free (backends);
	    free (values);
	    free (weights);
	    return;
	}

	/* ELSE: Choose a random one of the availables. */
	current_backend = backends[rand() % nbackends];
	servicereport->last_backend = current_backend;
	msg ("chosen backend (flat-weight random): %d", current_backend);
	free (backends);
	free (values);
	return;

    case ds_bysize:
	/* Fill the values */
	values = xmalloc (nbackends * sizeof(double));
	for (i = 0; i < nbackends; i++) {
	    /* When dispatch over #-connections is in place, then we
	     * compute using the averaged value. However if the averaged
	     * value is 0 (not yet computed), we take the true value. */
	    if (activeservice->dispatchover &&
		servicereport->backendstate[backends[i]].avg_nbytes)
		values[i] =
		    servicereport->backendstate[backends[i]].avg_nbytes *
		    activeservice->backend[backends[i]].weight;
	    else
		values[i] = servicereport->backendstate[backends[i]].nbytes *
		    activeservice->backend[backends[i]].weight;
	    msg ("by size weighing backend %d has weight %d, value %u"
		 " (bytes=%u, avgbytes=%u)",
		 backends[i], 
		 activeservice->backend[backends[i]].weight,
		 (unsigned) values[i],
		 (unsigned) servicereport->backendstate[backends[i]].nbytes,
		 (unsigned)
		     servicereport->backendstate[backends[i]].avg_nbytes);
	}
	/* Get the backend that has transported the least bytes */
	for (i = 0; i < nbackends; i++)
	    if (!target_set++ || values[i] < nbest) {
		nbest = values[i];
		current_backend = backends[i];
		servicereport->last_backend = current_backend;
		msg ("by size weighing: best so far is %d (value %g)",
		     current_backend, nbest);
	    }
	msg ("chosen backend (by size): %d", current_backend);
	free (backends);
	free (values);
	return;

    case ds_byduration:
	/* Fill the values */
	values = xmalloc (nbackends * sizeof(double));
	for (i = 0; i < nbackends; i++) {
	    /* When dispatch over #-connections is in place, then we
	     * compute using the averaged value. However if the averaged
	     * value is 0 (not yet computed), we take the true value. */
	    if (activeservice->dispatchover &&
		servicereport->backendstate[backends[i]].avg_nsec)
		values[i] =
		    servicereport->backendstate[backends[i]].avg_nsec *
		    1000000 *
		    activeservice->backend[backends[i]].weight;
	    else
		values[i] = servicereport->backendstate[backends[i]].nsec *
		    1000000 *
		    activeservice->backend[backends[i]].weight;
	    msg ("By duration weighing: backend %d has value %g"
		 " (sec=%g, avgsec=%g, weight=%d)",
		 backends[i], values[i],
		 servicereport->backendstate[backends[i]].nsec,
		 servicereport->backendstate[backends[i]].avg_nsec,
		 activeservice->backend[backends[i]].weight);
	}
	/* Get the backend that has transported the least connection secs */
	for (i = 0; i < nbackends; i++)
	    if (!target_set++ || values[i] < nbest) {
		nbest = values[i];
		current_backend = backends[i];
		servicereport->last_backend = current_backend;
		msg ("by duration weighing: best so far is %d (value %g)",
		     current_backend, nbest);
	    }
	msg ("chosen backend (by duration): %d", current_backend);
	free (backends);
	free (values);
	return;

    case ds_byorder:
	/* Get the first available back end in line. */
	current_backend = backends[0];
	servicereport->last_backend = current_backend;
	msg ("chosen backend (by order): %d", current_backend);
	free (backends);
	return;

    case ds_byconnections:
    case ds_byclientip:
	/* Note: this serves byconnections dispatching, plus byclientip
	 * when the actual back end has gone down. */
	values = xmalloc (nbackends * sizeof(double));
	for (i = 0; i < nbackends; i++) {
	    values[i] =
		servicereport->backendstate[backends[i]].nclients *
		activeservice->backend[backends[i]].weight;
	    msg ("by connections weighing: backend %d has value %g",
		 backends[i], values[i]);
	}
	/* Find the minimum in the values array. */
	for (i = 0; i < nbackends; i++)
	    if (!target_set++ || values[i] < nclients) {
		nclients = values[i];
		current_backend = backends[i];
		servicereport->last_backend = current_backend;
		msg ("by connections weighing: best so far is %d (value %u)",
		     current_backend, nclients);
	    }
	msg ("chosen backend (by connections): %d", current_backend);
	free (backends);
	free (values);
	return;

    case ds_externalhandler:
	/* External handler to be called. Assume selection will fail. */
	current_backend = -1;
	exthandler = str_expand_format (activeservice->dispatchext);
	msg ("calling external handler '%s'", exthandler);
	signal (SIGCHLD, SIG_DFL);
	uid_assume();
	if (! (f = popen (exthandler, "r")) ) {
	    warning ("failed to start external handler '%s': %s",
		     exthandler, strerror(errno));
	    if (nbackends > 0) {
		current_backend = backends[0];
		servicereport->last_backend = current_backend;
	    }
	    
	    uid_restore();
	    signal (SIGCHLD, SIG_IGN);
	    free (backends);
	    free (exthandler);
	    return;
	}
	while (1) {
	    if (fscanf (f, " %80s ", buf) < 1) {
		msg ("external handler signals end");
		pclose (f);
		uid_restore();
		signal (SIGCHLD, SIG_IGN);
		break;
	    }
	    msg ("external handler says '%s'", buf);
	    for (i = 0; i < activeservice->nbackend; i++) {
		if (!strcmp (buf, activeservice->backend[i].name)) {
		    msg ("selecting back end %s due to external handler", buf);
		    current_backend = i;
		    servicereport->last_backend = current_backend;
		    pclose (f);
		    uid_restore();
		    signal (SIGCHLD, SIG_IGN);
		    free (backends);
		    free (exthandler);
		    return;
		}
	    }
	}
	warning ("external handler '%s' didn't reply with a selectable back end",
		 exthandler);
	if (nbackends > 0)
	    current_backend = backends[0];
	servicereport->last_backend = current_backend;
	free (backends);
	free (exthandler);
	return;

    default:
	/* Internal fry.. */
	error ("internal error: unhandled dispatch type %d",
	       activeservice->dispatchtype);
    }
}
