/*************************************************************************
 * This file is part of Crosroads 1.79, a load balancer and fail over
 * utility for TCP. Copyright (c) Karel Kubat, distributed under GPL.
 * Visit http://crossroads.e-tunity.com for information.
 *************************************************************************/

#include "../crossroads.h"

static void rd (int fd, void *buf, unsigned sz) {
    int res;
    int toread = (int) sz;

    if ( (res = read (fd, buf, sz)) != toread )
	error ("Failed to fetch %d bytes (%d)", toread, res);
}

static int rd_int (int fd) {
    int res;

    rd (fd, &res, sizeof(int));
    return (res);
}

static int rd_uns (int fd) {
    unsigned res;

    rd (fd, &res, sizeof(unsigned));
    return (res);
}

static char *rd_str (int fd) {
    char *buf;
    int len;

    len = rd_int (fd);
    if (!len)
	return (0);
    if (len > 1024)
	error ("Maximum string length exceeded, attempt for %d", len);
    buf = xcalloc (len + 1);
    rd (fd, buf, len);
    return (buf);
}


void config_read (int fd) {
    int i, j, k;

    msg ("Reading binary service definition from fd %d", fd);

    nservice = rd_int (fd);
    msg ("Configured services: %d", nservice);
    service = xcalloc (nservice * sizeof(Service));
    for (i = 0; i < nservice; i++) {
	/* Basic stuff */
	service[i].name = rd_str (fd);
	msg ("  Scanning service %d (%s)", i + 1, service[i].name);
	service[i].bind = rd_str (fd);
	service[i].port = rd_uns (fd);
	service[i].verbosity = rd_uns (fd);
	service[i].dispatchtype = (Dispatchtype)rd_int (fd);
	service[i].dispatchover = rd_uns (fd);
	service[i].dispatchext = rd_str (fd);
	service[i].rev_interval = rd_uns (fd);
	service[i].check_interval = rd_uns (fd);
	service[i].check_cmd = rd_str (fd);
	service[i].backlog = rd_uns (fd);
	service[i].shmkey = rd_uns (fd);
	service[i].connectiontimeout = rd_uns (fd);
	service[i].maxconnections = rd_uns (fd);
	service[i].type = (Dispatchtype)rd_int (fd);
	service[i].allowfile = rd_str (fd);
	service[i].denyfile = rd_str (fd);
	service[i].uid = rd_int (fd);
	service[i].gid = rd_int (fd);
	service[i].inspection = rd_int (fd);

	/* Deny chain */
	service[i].ndenychain = rd_int (fd);
	service[i].denychain = xcalloc (service[i].ndenychain *
					sizeof(IpFilter));
	for (j = 0; j < service[i].ndenychain; j++) {
	    service[i].denychain[j].ip = rd_int (fd);
	    service[i].denychain[j].mask = rd_int (fd);
	}
    
	/* Allow chain */
	service[i].nallowchain = rd_int (fd);
	service[i].allowchain = xcalloc (service[i].nallowchain *
					sizeof(IpFilter));
	for (j = 0; j < service[i].nallowchain; j++) {
	    service[i].allowchain[j].ip = rd_int (fd);
	    service[i].allowchain[j].mask = rd_int (fd);
	}

	/* Back ends */
	service[i].nbackend = rd_int (fd);
	msg ("  Configured back ends: %d", service[i].nbackend);
	service[i].backend = xcalloc (service[i].nbackend *
				      sizeof(Backend));
	for (j = 0; j < service[i].nbackend; j++) {
	    service[i].backend[j].name = rd_str (fd);
	    msg ("    Backend %d: %s", j + i, service[i].backend[j].name);
	    service[i].backend[j].initial_port = rd_int (fd);
	    service[i].backend[j].initial_server = rd_str (fd);
	    service[i].backend[j].verbosity = rd_int (fd);
	    service[i].backend[j].weight = rd_int (fd);
	    service[i].backend[j].decay = rd_int (fd);
	    service[i].backend[j].onstart = rd_str (fd);
	    service[i].backend[j].onend = rd_str (fd);
	    service[i].backend[j].dumpfile = rd_str (fd);
	    service[i].backend[j].thruputfile = rd_str (fd);
	    service[i].backend[j].timinglog = rd_str (fd);
	    service[i].backend[j].maxconnections = rd_uns (fd);
	    service[i].backend[j].stickycookie = rd_str (fd);
	    service[i].backend[j].retries = rd_int (fd);
	    service[i].backend[j].initial_avail = rd_int (fd);

	    service[i].backend[j].naddclientheader = rd_int (fd);
	    service[i].backend[j].addclientheader =
		xcalloc (service[i].backend[j].naddclientheader *
			 sizeof(char*));
	    for (k = 0; k < service[i].backend[j].naddclientheader; k++)
		service[i].backend[j].addclientheader[k] = rd_str (fd);

	    service[i].backend[j].nsetclientheader = rd_int (fd);
	    service[i].backend[j].setclientheader =
		xcalloc (service[i].backend[j].nsetclientheader *
			 sizeof(char*));
	    for (k = 0; k < service[i].backend[j].nsetclientheader; k++)
		service[i].backend[j].setclientheader[k] = rd_str (fd);

	    service[i].backend[j].nappendclientheader = rd_int (fd);
	    service[i].backend[j].appendclientheader =
		xcalloc (service[i].backend[j].nappendclientheader *
			 sizeof(char*));
	    for (k = 0; k < service[i].backend[j].nappendclientheader; k++)
		service[i].backend[j].appendclientheader[k] = rd_str (fd);

	    service[i].backend[j].naddserverheader = rd_int (fd);
	    service[i].backend[j].addserverheader =
		xcalloc (service[i].backend[j].naddserverheader *
			 sizeof(char*));
	    for (k = 0; k < service[i].backend[j].naddserverheader; k++)
		service[i].backend[j].addserverheader[k] = rd_str (fd);

	    service[i].backend[j].nsetserverheader = rd_int (fd);
	    service[i].backend[j].setserverheader =
		xcalloc (service[i].backend[j].nsetserverheader *
			 sizeof(char*));
	    for (k = 0; k < service[i].backend[j].nsetserverheader; k++)
		service[i].backend[j].setserverheader[k] = rd_str (fd);

	    service[i].backend[j].nappendserverheader = rd_int (fd);
	    service[i].backend[j].appendserverheader =
		xcalloc (service[i].backend[j].nappendserverheader *
			 sizeof(char*));
	    for (k = 0; k < service[i].backend[j].nappendserverheader; k++)
		service[i].backend[j].appendserverheader[k] = rd_str (fd);
	}
    }

    msg ("Configuration scanned.");
}
