/*************************************************************************
 * This file is part of Crosroads 1.79, a load balancer and fail over
 * utility for TCP. Copyright (c) Karel Kubat, distributed under GPL.
 * Visit http://crossroads.e-tunity.com for information.
 *************************************************************************/

#include "../crossroads.h"

/* For debugging, uncomment: */
#define DEBUG

static void wr (int fd, void const *buf, unsigned len) {
    int res;
    int towrite = (int)len;
    
    if ( (res = write (fd, buf, len)) != towrite )
	 error ("Failed to transmit %d bytes (%d)", towrite, res);
}

static void wr_int (int fd, int x, char const *desc) {
    #ifdef DEBUG
    msg ("Configwrite: int %d (%u bytes), %s", x, (unsigned)sizeof(int), desc);
    #endif
    wr (fd, &x, sizeof(int));
}

static void wr_uns (int fd, unsigned x, char const *desc) {
    #ifdef DEBUG
    msg ("Configwrite: unsigned %u (%u bytes), %s", x, (unsigned)sizeof(int),
	 desc);
    #endif
    wr (fd, &x, sizeof(unsigned));
}

static void wr_str (int fd, char const *s, char const *desc) {
    if (!s || !*s) {
	#ifdef DEBUG
	msg ("Configwrite: empty string, %s", desc);
	#endif
	wr_int (fd, 0, "zero string length");
    } else {
	#ifdef DEBUG
	msg ("Configwrite: string %s (%u bytes), %s", s, (unsigned)strlen(s), desc);
	#endif
	wr_int (fd, strlen(s), "nonzero string length");
	wr (fd, s, strlen(s));
    }
}

void config_write (int fd) {
    int i, j, k;

    msg ("Writing binary service definition to fd %d", fd);
    wr_int (fd, nservice, "number of services");
    for (i = 0; i < nservice; i++) {	
	/* Basic stuff */
	wr_str (fd, service[i].name, "servicename");
	wr_str (fd, service[i].bind, "servicebind");
	wr_uns (fd, service[i].port, "serviceport");
	wr_uns (fd, service[i].verbosity, "serviceverbosity");
	wr_int (fd, (int)service[i].dispatchtype, "dispatchtype");
	wr_uns (fd, service[i].dispatchover, "dispatchover");
	wr_str (fd, service[i].dispatchext, "dispatchext");
	wr_uns (fd, service[i].rev_interval, "revivinginterval");
	wr_uns (fd, service[i].check_interval, "checkinterval");
	wr_str (fd, service[i].check_cmd, "checkcommand");
	wr_uns (fd, service[i].backlog, "backlog");
	wr_uns (fd, service[i].shmkey, "shmkey");
	wr_uns (fd, service[i].connectiontimeout, "connectiontimeout");
	wr_uns (fd, service[i].maxconnections, "maxconnections");
	wr_int (fd, (int)service[i].type, "type");
	wr_str (fd, service[i].allowfile, "allowfile");
	wr_str (fd, service[i].denyfile, "denyfile");
	wr_int (fd, service[i].uid, "uid");
	wr_int (fd, service[i].gid, "gid");
	wr_int (fd, service[i].inspection, "inspectiontype");

	/* Deny chain */
	wr_int (fd, service[i].ndenychain, "sizeof denychain");
	for (j = 0; j < service[i].ndenychain; j++) {
	    wr_uns (fd, service[i].denychain[j].ip, "denied ip");
	    wr_uns (fd, service[i].denychain[j].mask, "denied mask");
	}

	/* Allow chain */
	wr_int (fd, service[i].nallowchain, "sizeof allowchain");
	for (j = 0; j < service[i].nallowchain; j++) {
	    wr_uns (fd, service[i].allowchain[j].ip, "allowed ip");
	    wr_uns (fd, service[i].allowchain[j].mask, "allowed mask");
	}

	/* Back ends */
	wr_int (fd, service[i].nbackend, "nr of backends");
	for (j = 0; j < service[i].nbackend; j++) {
	    wr_str (fd, service[i].backend[j].name, "backend name");
	    wr_int (fd, service[i].backend[j].initial_port, "backend port");
	    wr_str (fd, service[i].backend[j].initial_server, "backend server");
	    wr_int (fd, service[i].backend[j].verbosity, "backend verbosity");
	    wr_int (fd, service[i].backend[j].weight, "backend weight");
	    wr_int (fd, service[i].backend[j].decay, "backend decay");
	    wr_str (fd, service[i].backend[j].onstart, "backend onstart");
	    wr_str (fd, service[i].backend[j].onend, "backend onend");
	    wr_str (fd, service[i].backend[j].dumpfile, "backend dumpfile");
	    wr_str (fd, service[i].backend[j].thruputfile,
		    "backend thruputfile");
	    wr_str (fd, service[i].backend[j].timinglog, "backend timinglog");
	    wr_uns (fd, service[i].backend[j].maxconnections,
		    "backend maxcconnections");
	    wr_str (fd, service[i].backend[j].stickycookie,
		    "backend stickycookie");
	    wr_int (fd, service[i].backend[j].retries, "backend retries");
	    wr_int (fd, service[i].backend[j].initial_avail,
		    "backend availability");

	    wr_int (fd, service[i].backend[j].naddclientheader,
		    "nr of clientheaders to add");
	    for (k = 0; k < service[i].backend[j].naddclientheader; k++)
		wr_str (fd, service[i].backend[j].addclientheader[k],
		    "client header to add");
	    
	    wr_int (fd, service[i].backend[j].nsetclientheader,
		    "nr of client headers to set");
	    for (k = 0; k < service[i].backend[j].nsetclientheader; k++)
		wr_str (fd, service[i].backend[j].setclientheader[k],
			"client header to set");
	    
	    wr_int (fd, service[i].backend[j].nappendclientheader,
		    "nr of client headers to append");
	    for (k = 0; k < service[i].backend[j].nappendclientheader; k++)
		wr_str (fd, service[i].backend[j].appendclientheader[k],
			"client header to append");

	    wr_int (fd, service[i].backend[j].naddserverheader,
		    "nr of server headers to add");
	    for (k = 0; k < service[i].backend[j].naddserverheader; k++)
		wr_str (fd, service[i].backend[j].addserverheader[k],
			"server header to add");
	    
	    wr_int (fd, service[i].backend[j].nsetserverheader,
		    "nr of server headers to set");
	    for (k = 0; k < service[i].backend[j].nsetserverheader; k++)
		wr_str (fd, service[i].backend[j].setserverheader[k],
			"server header to set");
	    
	    wr_int (fd, service[i].backend[j].nappendserverheader,
		    "nr of server headers to append");
	    for (k = 0; k < service[i].backend[j].nappendserverheader; k++)
		wr_str (fd, service[i].backend[j].appendserverheader[k],
			"server header to append");
	}
    }
}
