/*************************************************************************
 * This file is part of Crosroads 1.79, a load balancer and fail over
 * utility for TCP. Copyright (c) Karel Kubat, distributed under GPL.
 * Visit http://crossroads.e-tunity.com for information.
 *************************************************************************/

#include "../crossroads.h"

/* #define DEBUG */

static int noncache_resolve (char const *hostname, struct in_addr *res) {
    struct hostent *hent;
    double tick;

    tick = timeofday();
    if (! (hent = gethostbyname(hostname)) )
	return (0);
    *res = *(struct in_addr *)hent->h_addr;
    http_log_activity (tick, 0, "hostname resolved");
    return (1);
}

struct in_addr *dns (char const *hostname) {
    int i, target = -1;
    time_t now, age, min;
 
    /* No cache TTL? No caching. We use slot 0 for storage of the
     * intermediate result. */
    if (!dns_cachettl) {
	if (noncache_resolve (hostname, &dnscache[0].hostaddr))
	    return (&dnscache[0].hostaddr);
	return (0);
    }

    /* Current time */
    now = time (0);
    
    #ifdef DEBUG
    if (program_stage != stage_retrying) {
	msg ("dnscache: now %d", now);
	for (i = 0; i < DNS_CACHESIZE; i++)
	    msg ("dnscache[%d].hostname: %s, dnscache[%d].stamp: %d",
		 i, dnscache[i].hostname, i, (int)dnscache[i].stamp);
    }
    #endif

    /* See if we still have the hostname. If so, if the TTL isn't
     * exceeded, return that. */
    for (i = 0; i < DNS_CACHESIZE && dnscache[i].hostname; i++) {
	if (!strcmp (dnscache[i].hostname, hostname)) {
	    age = now - dnscache[i].stamp;
	    if (age <= dns_cachettl) {
		if (program_stage != stage_retrying)
		    msg ("host %s resolved from cache slot %d (age: %d sec)", 
			 hostname, i, (int)age);
		return (&(dnscache[i].hostaddr));
	    } else {
		if (program_stage != stage_retrying)
		    msg ("host entry for %s is too old at slot %d "
			 "(age: %d sec, ttl %d sec)",
			 hostname, i, (int)age,
			 dns_cachettl);
		target = i;
		break;
	    }	
	}
    }

    /* We failed to return from cache. If target isn't set yet, set it. */
    if (target == -1) {
	if (i < DNS_CACHESIZE) {
	    target = i;
	    if (program_stage != stage_retrying)
		msg ("host entry '%s' at slot %d will be added",
		     hostname, target);
	} else {
	    for (i = 0; i < DNS_CACHESIZE; i++)
		if (i == 0 || dnscache[i].stamp < min) {
		    target = i;
		    min = dnscache[i].stamp;
		    
		}
	    if (program_stage != stage_retrying)
		msg ("host entry '%s' at slot %d "
		     "will be overwritten",
		     hostname, target);
	}	
    }

    /* Get the name, and stick it in the cache if it's a resolved entry,
     * return the result.*/
    if (noncache_resolve (hostname, &dnscache[i].hostaddr)) {
	free (dnscache[i].hostname);
	dnscache[i].hostname = xstrdup (hostname);
	dnscache[i].stamp = now;
	return (&dnscache[i].hostaddr);
    }	

    /* Failed, return a NULL ptr. */
    return (0);
}
