/*************************************************************************
 * This file is part of Crosroads 1.79, a load balancer and fail over
 * utility for TCP. Copyright (c) Karel Kubat, distributed under GPL.
 * Visit http://crossroads.e-tunity.com for information.
 *************************************************************************/
#include "../crossroads.h"

static void copy (int src, int dst, int dir, int tot) {
    int ncopied = 0, to_copy, bytes;
    unsigned char *buf;

    msg ("about to copy %d bytes from %s", 
	 tot, dir == dir_client_to_server ? "client" : "server");
    msg ("%d bytes in client buffer, %d in server buffer",
	 clbufmax ? clbufmax - clbufpos : 0,
	 srbufmax ? srbufmax - srbufpos : 0);

    while (ncopied < tot) {
	/* Here's how many bytes we should shuttle to and fro.
	 * This is the remaining count, or if the buffer is too small
	 * for that, the buffer size. */
	to_copy = tot - ncopied;
	if (to_copy > opt.tcp_bufsz)
	    to_copy = opt.tcp_bufsz;

	buf = net_bufread (src, to_copy, &bytes, dir == dir_client_to_server);
	net_write (dst, buf, bytes, dir == dir_client_to_server);
	ncopied += bytes;

	/* msg ("Copied 0x%x (%d) bytes of 0x%x (%d) total",
	     ncopied, ncopied, tot, tot);
	*/
    }
}

unsigned getchunk (int sock, int dest, CopyDirection dir) {
    unsigned ret = 0;
    unsigned char *cp;
    char *hexbuf = 0;
    int in_number = 1;

    /* We match
     *           <hex-chunksize> <any-mush> [\r]\n
     * while copying all to the other socket, and return the
     * chunk size */
    while (1) {
	cp = net_bufread (sock, 1, 0, dir == dir_client_to_server);
	/* msg ("Got character %c (%d) while scanning chunk size",
	    *cp, *cp); */
	net_write (dest, cp, 1, dir == dir_client_to_server);
	if (*cp == '\n') {
	    sscanf (hexbuf, "%x", &ret);
	    free (hexbuf);
	    msg ("Chunk size 0x%x (%d)", ret, ret);
	    /* If the chunk size is zero, then match final [\r]\n */
	    if (ret == 0) {
		cp = net_bufread (sock, 1, 0, dir == dir_client_to_server);
		net_write (dest, cp, 1, dir == dir_client_to_server);
		if (*cp == '\r') {
		    cp = net_bufread (sock, 1, 0, dir == dir_client_to_server);
		    net_write (dest, cp, 1, dir == dir_client_to_server);
		    if (*cp != '\n')
			warning ("Chunks not terminated by \\r\\n "
				 "but by \\r <decimal>%d", *cp);
		} else if (*cp != '\n')
		    warning ("Chunks not terminated by \r\n or \n "
			     "but by <decimal>%d", *cp);
	    }
	    return (ret);
	}
	if (in_number) {
	    if (is_hex_digit (*cp))
		hexbuf = xstrcatch (hexbuf, *cp);
	    else
		in_number = 0;
	}
    }
}

void http_copy (HttpHeader *h, int src, int dst, CopyDirection dir) {
    unsigned char const *mode, *cp;
    unsigned nbytes;
    
    msg ("copying HTTP body from %s", 
	 dir == dir_client_to_server ?
	   "client to server" : "server to client");

    /* Check in what 'mode' we are.
     * We know 3 modes:
     * - Transfer-Encoding is chunked: we copy chunks
     * - Content-Length is supplied: we copy that # of bytes
     * - None of the above: we don't know, go to copy-thru mode
     */
  
    if ( (mode = http_header_val (h, "transfer-encoding")) && *mode ) {
	if (strncasecmp ( (char const *) mode, "chunked", 7))
	    error ("can't handle transfer-encoding '%s'", mode);
	msg ("copying chunks from %s",
	     dir == dir_client_to_server ?
	         "client to server" : "server to client");
	while (1) {
	    /* See how large the chunk will be. */
	    nbytes = getchunk (src, dst, dir);
	    msg ("Next chunk is 0x%x (%d) bytes", nbytes, nbytes);
	    /* Copy the chunk. Add 2 bytes for the \r\n that follows. */
	    copy (src, dst, dir, nbytes + 2);
	    /* Last chunksize of 0 signals the end. */
	    if (!nbytes)
		break;
	}
    } else if ( (mode = http_header_val (h, "content-length")) &&
		mode ) {
	nbytes = atoi ( (char const *) mode );
	msg ("copying %u bytes of content from %s",
	     nbytes, 
	     dir == dir_client_to_server ?
	         "client to server" : "server to client");
	copy (src, dst, dir, nbytes);
    } else if (dir == dir_server_to_client) {
	msg ("Starting copy-thru on body sending towards client");
	if ( (cp = net_buffer (dir_server_to_client, &nbytes)) ) {
	    msg ("Flushing server-side buffer to client, %u bytes", nbytes);
	    net_write (dst, cp, nbytes, 0);
	}
	copysockets (src, dst);
    } else {
	msg ("Not copying body (nothing to do)");
    }
}
