/*************************************************************************
 * This file is part of Crosroads 1.79, a load balancer and fail over
 * utility for TCP. Copyright (c) Karel Kubat, distributed under GPL.
 * Visit http://crossroads.e-tunity.com for information.
 *************************************************************************/
#include "../crossroads.h"

void http_serve (int clientsock) {
    int serversock = -1, is_continuation, i;
    HttpHeader *clientreq, *serverresp;
    int client_persisting, server_persisting;
    double cl_header_start, cl_header_end,
	con_start, con_end, start, tcptotal, tick;

    /* Set relevant sigs etc. */
    program_stage = stage_serving;
    for (i = 0; relevant_sigs[i]; i++)
	signal (relevant_sigs[i], interrupt);
    
    set_program_title ("crossroads - Service %s: serving %s",
		       activeservice->name, client_ip);   
    
    /* We're the child branch now of fork_tcp_servicer().
     * Here's the logic of the HTTP services.
     * - Get the client's request;
     * - If we're not yet connected to a back end: determine the back
     *   end and connect to it
     * - Apply all server-directed headers
     * - Send client's request to the server
     * - Read back the server's response
     * - Apply all client-directed headers
     * - Send the server's response to the client
     * - Redo from start; this might be a Keep-Alive connection.
     *   Alternatively, if deep inspection is off, go to cpio-thru mode.
     */

    net_allocbufs();
    while (1) {
	start = timeofday();
	tcptotal = 0;
	/* Reset network IO buffers. */
	clbufpos = clbufmax = srbufpos = srbufmax = 0;
	
	/* Assume both client and server want to persist. */
	client_persisting = server_persisting = 1;
	
	/* Get the initial client headers. We store the times but do
	 * not log yet, that can be done once we assign a back end. */
	clientreq = http_header_new();
	cl_header_start = timeofday();
	http_header_read (clientreq, clientsock, dir_client_to_server);
	cl_header_end = timeofday();
	tcptotal += (cl_header_end - cl_header_start);
	
	/* Determine the back end if necessary. */
	if (serversock == -1) {
	    con_start = timeofday();
	    if ( (serversock = http_serversocket (clientreq,
						  &is_continuation)) < 1 )
		http_error (clientsock);
	    con_end = timeofday();
	    http_log_activity (cl_header_start, cl_header_end,
			       "client headers received");
	    http_log_activity (con_start, con_end, "connected to backend");
	    tcptotal += (con_end - con_start);	    
	    flag_verbose = activeservice->backend[current_backend].verbosity;
	    
	    incr_client_count();
	    if (is_continuation)
		log_activity_continuation();
	    else
		log_activity_start ();
	}
	
	/* Put all server-directed headers in place. */
	for (i = 0;
	     i < activeservice->backend[current_backend].naddserverheader;
	     i++)
	    http_header_addheader (clientreq,
	        activeservice->backend[current_backend].addserverheader[i]);
	for (i = 0;
	     i < activeservice->backend[current_backend].nsetserverheader;
	     i++)
	    http_header_setheader (clientreq,
  	        activeservice->backend[current_backend].setserverheader[i]);
	for (i = 0;
	     i < activeservice->backend[current_backend].nappendserverheader;
	     i++)
	    http_header_appendheader (clientreq,
	        activeservice->backend[current_backend].appendserverheader[i]);

	/* See RFC2616. We're actually a proxy here. Clients that talk
	 * HTTP/1.0 MUST be treated as single shot connection clients, even
	 * when they specify Connection: Keep-Alive. */
	if (http_header_httpver (clientreq) < 1.1) {
	    msg ("Client talks HTTP < 1.1, forcing closing connections");
	    http_header_setheader (clientreq, "Connection: close");
	    http_header_setheader (clientreq, "Proxy-Connection: close");
	    client_persisting = 0;
	} else if (http_header_connectiontype (clientreq) == con_close) {
	    msg ("Client asks for connection closing");
	    client_persisting = 0;
	}

	/* Send client's headers to the back end. */
	tick = timeofday();
	http_header_write (clientreq, serversock, dir_client_to_server);
	tcptotal += timeofday() - tick;
	http_log_activity (tick, 0, "sent clientheaders to backend");
	
	/* Copy body from client to server */
	tick = timeofday();
	http_copy (clientreq, clientsock, serversock, dir_client_to_server);
	tcptotal += timeofday() - tick;
	http_log_activity (tick, 0, "sent clientbody to backend");
	
	/* Get server response. */
	serverresp = http_header_new();
	tick = timeofday();
	http_header_read (serverresp, serversock, dir_server_to_client);
	tcptotal += timeofday() - tick;
	http_log_activity (tick, 0, "received headers from backend");

	/* Put client-directed headers in place. */
	for (i = 0;
	     i < activeservice->backend[current_backend].naddclientheader;
	     i++)
	    http_header_addheader (serverresp,
	        activeservice->backend[current_backend].addclientheader[i]);
	for (i = 0;
	     i < activeservice->backend[current_backend].nsetclientheader;
	     i++)
	    http_header_setheader (serverresp,
	        activeservice->backend[current_backend].setclientheader[i]);
	for (i = 0;
	     i < activeservice->backend[current_backend].nappendclientheader;
	     i++)
	    http_header_appendheader (serverresp,
	        activeservice->backend[current_backend].appendclientheader[i]);

	/* According to RFC1616, we should convert chunked output from
	 * the server to one big message. We won't do that here, it breaks
	 * streaming. But we will close connections in the case of chunked
	 * transfer. */
	if (client_persisting &&
	    http_header_val (serverresp, "transfer-encoding")) {
	    msg ("Server is sending chunks, forcing singleshot");
	    server_persisting = 0;
	    client_persisting = 0;
	    http_header_setheader (serverresp, "Connection: close");
	    http_header_setheader (serverresp, "Proxy-Connection: close");
	}

	/* If the server wants to close the connection, then we must
	 * inform the client. */
	if (client_persisting) {
	    if (http_header_httpver (serverresp) < 1.1) {
		msg ("Server talks HTTP < 1.1, forcing closing connections");
		http_header_setheader (serverresp, "Connection: close");
		http_header_setheader (serverresp, "Proxy-Connection: close");
		server_persisting = 0;
	    } else if (http_header_connectiontype (serverresp) == con_close) {
		msg ("Server asks for connection closing");
		http_header_setheader (serverresp, "Connection: close");
		http_header_setheader (serverresp, "Proxy-Connection: close");
		server_persisting = 0;
	    }
	}
	
	/* Send server headers to the client. */
	tick = timeofday();
	http_header_write (serverresp, clientsock, dir_server_to_client);
	tcptotal += timeofday() - tick;
	http_log_activity (tick, 0, "sent serverheaders to client");
	
	/* Copy body from server to client and flush network buffers if any */
	tick = timeofday();
	http_copy (serverresp, serversock, clientsock, dir_server_to_client);
	tcptotal += timeofday() - tick;
	http_log_activity (tick, 0, "sent serverbody to client");

	/* Free up info from this request/response chat. */
	http_header_free (clientreq);
	http_header_free (serverresp);

	/* Overall activity reporting */
	http_log_activity (start, 0, "total processing (%g in TCP)", tcptotal);

	/* Should we loop around? Go to copy-thru mode if either the
	 * client or the server wants to stop. */
	if (!client_persisting || !server_persisting) {
	    msg ("non-persistent connection");
	    break;
	}

	if (activeservice->inspection == ins_deep) 
	    msg ("Will analyze following requests over the same TCP link");
	else {
	    msg ("Shallow analysis - won't analyze following "
		 "requests of the same TCP link");
	    break;
	}
    }

    /* We're done analyzing. If there's leftovers in the network buffers,
     * then we flush them first. */
    if (clbufmax) {
	msg ("Flushing client-side buffer to server, %u bytes",
	     clbufmax - clbufpos);
	http_write (serversock, dir_client_to_server,
		    clbuf + clbufpos, clbufmax - clbufpos);
	clbufmax = clbufpos = 0;
    }
    if (srbufmax) {
	msg ("Flushing server-side buffer to client, %u bytes",
	     srbufmax - srbufpos);
	http_write (clientsock, dir_server_to_client,
		    srbuf + srbufpos, srbufmax - srbufpos);
	srbufmax = srbufpos = 0;
    }

    /* Copy through all that arrives on back end or client.
     * This doesn't return... */
    msg ("HTTP service done analyzing, entering copy-thru mode");

    /* This doesn't return */
    copysockets (clientsock, serversock);
}
