/*************************************************************************
 * This file is part of Crosroads 1.79, a load balancer and fail over
 * utility for TCP. Copyright (c) Karel Kubat, distributed under GPL.
 * Visit http://crossroads.e-tunity.com for information.
 *************************************************************************/

#include "../crossroads.h"

void mark_activity (double nbytes, double nsec,
		    Backendavail newstate) {
    double llval;
    double dlval;
    unsigned multiplier;
    int i, marked_unavailable = 0;

    /* If we had a signal somewhere, don't update the values.
     * If we don't have a back end yet, don't update the values. */
    if 	(interrupted || current_backend == -1)
	return;

    /* Update values. */
    lock_reporter();

    /* If the new activity is a failure: increase fails count.
     * Upon other state, set fails count to zero. */
    if (newstate == st_unavailable &&
	servicereport->backendstate[current_backend].actual_avail !=
	st_unavailable &&
	program_stage != stage_retrying) {
	servicereport->backendstate[current_backend].fail++;
	msg ("Back end %d's fail count increased to %d",
	     current_backend,
	     servicereport->backendstate[current_backend].fail);
    } else if (newstate == st_available)
	servicereport->backendstate[current_backend].fail = 0;
    
    if (program_stage != stage_retrying) {
	/* Handle the decays of all other back ends */
	for (i = 0; i < activeservice->nbackend; i++) {
	    if (i != current_backend && activeservice->backend[i].decay) {
		if (activeservice->dispatchover) {
		    servicereport->backendstate[i].avg_nbytes *=
			(100 - activeservice->backend[i].decay);
		    servicereport->backendstate[i].avg_nbytes /= 100;
		    servicereport->backendstate[i].avg_nsec *=
			(100 - activeservice->backend[i].decay);
		    servicereport->backendstate[i].avg_nsec /= 100;
		} else {
		    servicereport->backendstate[i].nbytes *=
			(100 - activeservice->backend[i].decay);
		    servicereport->backendstate[i].nbytes /= 100;
		    servicereport->backendstate[i].nsec *=
			(100 - activeservice->backend[i].decay);
		    servicereport->backendstate[i].nsec /= 100;
		}
	    }
	}

	/* Update average seconds and bytes. */
	if (activeservice->dispatchover) {
	    multiplier = activeservice->dispatchover - 1;
	    if (servicereport->backendstate[current_backend].totuses <
		multiplier)
		multiplier =
		    servicereport->backendstate[current_backend].totuses;

	    dlval = servicereport->backendstate[current_backend].avg_nsec;
	    dlval *= multiplier;
	    dlval += nsec;
	    dlval /= (multiplier + 1);
	    servicereport->backendstate[current_backend].avg_nsec = dlval;

	    llval = servicereport->backendstate[current_backend].avg_nbytes;
	    llval *= multiplier;
	    llval += nbytes;
	    llval /= (multiplier + 1);
	    servicereport->backendstate[current_backend].avg_nbytes = llval;
	}

	/* Update total secs / bytes. */
	servicereport->backendstate[current_backend].nsec += nsec;
	servicereport->backendstate[current_backend].nbytes += nbytes;
	if (newstate == st_unavailable)
	    servicereport->backendstate[current_backend].failures++;
    }
    
    /* Set the state, unless it's already marked for wakeup or brought down.
     * Do this only for 'final' states, not for intermediate. */
    if (servicereport->backendstate[current_backend].actual_avail
	!= st_waking &&
	servicereport->backendstate[current_backend].actual_avail
	!= st_down) {
	/* When trying to mark as unavailable, only proceed when we
	 * actually exceed the retries setting. */
	if (newstate == st_unavailable) {
	    if (servicereport->backendstate[current_backend].fail >=
		(unsigned) activeservice->backend[current_backend].retries) {
		msg ("backend %d: %d fails, %d retries, going offline",
		     current_backend,
		     servicereport->backendstate[current_backend].fail,
		     activeservice->backend[current_backend].retries);
		servicereport->backendstate[current_backend].fail = 0;
		servicereport->backendstate[current_backend].actual_avail =
		    st_unavailable;
		marked_unavailable++;
	    }
	    servicereport->backendstate[current_backend].totuses++;
	} else if (newstate != st_intermediate) {
	    /* When applying any other state than intermediate, it's OK */
	    servicereport->backendstate[current_backend].fail = 0;
	    servicereport->backendstate[current_backend].actual_avail =
		newstate;
	    servicereport->backendstate[current_backend].totuses++;
	}
    }

    /* Don't update servicereport beyond this point! */
    unlock_reporter();

    /* Give some feedback (though not during intermediate updates) */      
    if (newstate != st_intermediate && program_stage != stage_retrying)
	msg ("updated stats for backend %d (%s): "
	     "hits=%lu, "
	     "totfails=%lu, secs=%g, avgsecs=%g, "
	     "bytes=%g, avgbytes=%lu, state=%s",
	     current_backend, activeservice->backend[current_backend].name,
	     servicereport->backendstate[current_backend].totuses,
	     servicereport->backendstate[current_backend].failures,
	     servicereport->backendstate[current_backend].nsec,
	     servicereport->backendstate[current_backend].avg_nsec,
	     servicereport->backendstate[current_backend].nbytes,
	     servicereport->backendstate[current_backend].avg_nbytes,
	     state_to_string
	         (servicereport->backendstate[current_backend].actual_avail));

    /* Run the onfailure hook if one is specified. */
    if (marked_unavailable && program_stage != stage_retrying) {
	msg ("extcmd-onfail: "
	     " current_backend = %d, clients = %d, totclients = %d, cmd = %s",
	     current_backend,
	     servicereport->backendstate[current_backend].nclients,
	     servicereport->nclients, 
	     activeservice->backend[current_backend].onfail);

	sysrun (activeservice->backend[current_backend].onfail);
    }
}
