
/* Parser of crossroads configuration files */

%{
/* Prologue */
#include "../crossroads.h"

#define YYSTYPE Confsets

/* Static parser vars */
static int i;				/* Loop counter */
static Backend cur_backend;		/* Storage for a handled backend */
static Service cur_service;		/* Storage for a handled service */

/* Parser debugging related */
static void pmsg (char const *fmt, ...) {
    va_list args;

    if (parser_debug) {
	printf ("PARSER: ");
	va_start (args, fmt);
	vprintf (fmt, args);
    }
}

/* Conversion of yytext for symbol table */
#define SYMBOL (symtab_lookup((yytext)))

/* Error handler for yyparse() */
static int yyerror (char *msg) {
    char *sym;

    if (! yytext || ! *yytext)
	sym = "<end of file>";
    else if (isspace (*yytext))
	sym = "<white space>";
    else if (isprint (*yytext))
	sym = yytext;
    else
	sym = "<nonprintable>";
    
    error ("Parse error at %s, line %d, '%s': %s",
	   parser_filename(), yylineno + 1,
	   sym, yyerrmsg);
}

/* Store an encountered string */
static char *laststr;
static void setlaststr (char const *what) {
    free (laststr);
    laststr = xstrdup (what);
}

/* Store an encountered number */
static int lastnr;
static void setlastnr (char const *what) {
    if (what && *what && *what == '0')
	lastnr = strtol (what, 0, 8);
    else
	lastnr = atoi (what);
}

/* Store an encountered 'over' number */
static int lastovernr;
static void setlastovernr (char const *what) {
    if (what && *what && *what == '0')
	lastovernr = strtol (what, 0, 8);
    else
	lastovernr = atoi(what);
}

/* Store an encountered 'externalhandler' */
static char *lastext;
static void setlastext (char const *what) {
    free (lastext);
    lastext = xstrdup (what);
}

/* Get the server part from HOSTNAME:PORT in allocated memory */
static char *serverpart (char const *what) {
    char *ret = xstrdup (what);
    char *cp;

    if ( (cp = strchr (ret, ':')) )
	*cp = 0;
    return (ret);
}

/* Get the port part from HOSTNAME:PORT */
static int portpart (char const *what) {
    char *cp;

    if ( (cp = strchr (what, ':')) )
	return (atoi (cp + 1));
    return (0);
}

/* Add a list of IP filters to the allowlist or the denylist */
static void add_any (char *what, int chain) {
    char *item, *last = (char*) add_any;
    int result;

    for (item = strtok_r (what, " ", &last);
	 item;
	 item = strtok_r (0, " ", &last)) {
	if (chain)
	    result = ipf_add_allow (&cur_service, item);
	else
	    result = ipf_add_deny  (&cur_service, item);
	if (result)
	    error ("Bad IP filter specifier '%s' at line %d",
		   item, yylineno + 1);
    }
}
static void add_allowfrom (char *what) {
    add_any (what, 1);
}
static void add_denyfrom (char *what) {
    add_any (what, 0);
}

/* Set uid/gid for external commands. */
static void setuseraccount (char *username) {
    struct passwd *pw;

    if (! (pw = getpwnam (username)) )
	error ("Invalid username '%s' at line %d", username, yylineno + 1);
    cur_service.uid = pw->pw_uid;
    cur_service.gid = pw->pw_gid;
}

%}

/* Tokens: the meta tokens */
%token IDENTIFIER NUMBER STRING

 /* Tokens: the literal ones (these go automatically into the docs) */
%token SERVICE PORT VERBOSITY MAXCONNECTIONS
       TYPE ANY HTTP
       BACKEND SERVER BINDTO CONNECTIONTIMEOUT
       ON OFF
       DISPATCHMODE ROUNDROBIN RANDOM BYDURATION BYSIZE BYCONNECTIONS
       BYORDER BYCLIENTIP
       OVER DECAY  
       REVIVINGINTERVAL CHECKINTERVAL RETRIES SHMKEY WEIGHT
       ONSTART ONFAIL BACKLOG 
       THROUGHPUTLOG TRAFFICLOG HTTPTIMINGLOG
       STICKYCOOKIE ADDCLIENTHEADER SETCLIENTHEADER APPENDCLIENTHEADER
       ADDSERVERHEADER SETSERVERHEADER APPENDSERVERHEADER
       ALLOWFROM DENYFROM ALLOWFILE DENYFILE
       EXTERNALHANDLER USERACCOUNT ONEND
       HEADERINSPECTION DEEP SHALLOW
       STATE AVAILABLE UNAVAILABLE DOWN
       OPTIONS LOGACTIVITY TCPBUFFERSIZE DNSCACHETTL LOGFACILITY
       SHMPERMISSIONS SLOPPYPORTBIND LEAVEPROCESSTITLE 

%%
/* Config file grammar rules */

input:
	input
        element
|
	element
;

element:
	service_element
|
	option_element
;

option_element:
	OPTIONS
	openbrace
	option_statements
	closebrace
;

option_statements:
	option_statements
	option_statement
|
	option_statement
;

option_statement:
	option_statement_expected
	optstat
;

optstat:
	logactivity_statement
|
	tcpbuffersize_statement
|
	dnscachettl_statement
|
	logfacility_statement
|
	shmpermissions_statement
|
	sloppyportbind_statement
|
	leaveprocesstitle_statement
;

logactivity_statement:
	LOGACTIVITY
	onoff_expected
	onoff
	semicol {
	    opt.log_activity = lastnr;
	}
;

tcpbuffersize_statement:
	TCPBUFFERSIZE
	number
	semicol {
	    opt.tcp_bufsz = lastnr;
	}
;

dnscachettl_statement:
	DNSCACHETTL
	number
	semicol {
	    opt.dns_cachettl = lastnr;
	}
;

logfacility_statement:
	LOGFACILITY
	number
	semicol {
	    opt.log_facility = lastnr;
	}
;

shmpermissions_statement:
	SHMPERMISSIONS
	number
	semicol {
	    opt.shmperm = lastnr;
	}
;	    

sloppyportbind_statement:
	SLOPPYPORTBIND
	onoff_expected
	onoff
	semicol {
	    opt.sloppyportbind = lastnr;
	}
;

leaveprocesstitle_statement:
	LEAVEPROCESSTITLE
	onoff_expected
	onoff
	semicol {
	    opt.leave_proctitle = lastnr;
	}
;

service_element:
	service
	servicename
	openbrace
	servicestatements
	closebrace {
	    /* Check integrity, supply defaults */
	    check_service (cur_service);
	    
	    if (!cur_service.shmkey)
		cur_service.shmkey = (unsigned) hashpjw (cur_service.name);
	    msg ("Service %s: shared memory key set to 0x%8.8x",
		 cur_service.name, cur_service.shmkey);

	    if (!cur_service.check_cmd)
		cur_service.check_cmd = xstrdup ("");
	    
	    /* Add to the list, reset cur_service for next run */
	    service = xrealloc (service, ++nservice * sizeof(Service));
	    service[nservice - 1] = cur_service;
	    memset (&cur_service, 0, sizeof(Service));
	}
;

service:
	service_expected
	SERVICE
;

servicename:
	servicename_expected
	IDENTIFIER {
	    pmsg ("service [%s]\n", SYMBOL);
	    for (i = 0; i < nservice; i++)
		if (!strcmp (service[i].name, SYMBOL))
		    error ("Duplicate definition of service '%s'", SYMBOL);
	    cur_service.name = xstrdup(SYMBOL);
	}
;

servicestatements:
	servicestatements
	servicestatement
|
	servicestatement
;

servicestatement:
	servicebody_expected
	servicebody
;

servicebody:
	portstatement {
	    pmsg ("sevice port [%d]\n", $1.set[0].v.ival);
	    cur_service.port = $1.set[0].v.ival;
	    free ($1.set);
	}
|
	bindstatement {
	    pmsg ("service binding [%s]\n", $1.set[0].v.sval);
	    cur_service.bind = $1.set[0].v.sval;
	    free ($1.set);
	}
|
	verbositystatement {
	    pmsg ("service verbosity [%d]\n", $1.set[0].v.ival);
	    cur_service.verbosity = $1.set[0].v.ival;
	    free ($1.set);
	}
|
	dispatchmodestatement {
	    pmsg ("service dispatch mode [%d]\n", $1.set[0].v.ival);
	    pmsg ("service dispatch over [%d]\n", lastovernr);
	    pmsg ("service dispatch exthandler [%s]\n", lastext);
	    cur_service.dispatchtype = $1.set[0].v.ival;
	    cur_service.dispatchover = lastovernr;
	    cur_service.dispatchext  = xstrdup(lastext);
	    free ($1.set);
	}
|
	revivingintervalstatement {
	    pmsg ("service revival interval [%d]\n", $1.set[0].v.ival);
	    pmsg ("service revival cmd [%s]\n", $1.set[1].v.sval);
	    cur_service.rev_interval = $1.set[0].v.ival;
	    cur_service.check_cmd = $1.set[1].v.sval;
	    free ($1.set);
	}
|
	checkintervalstatement {
	    pmsg ("service check interval [%d]\n", $1.set[0].v.ival);
	    pmsg ("service check cmd [%s]\n", $1.set[1].v.sval);
	    cur_service.check_interval = $1.set[0].v.ival;
	    cur_service.check_cmd = $1.set[1].v.sval;
	}
|
	backlogstatement {
	    pmsg ("service backlog [%d]\n", $1.set[0].v.ival);
	    cur_service.backlog = $1.set[0].v.ival;
	    free ($1.set);
	}
|
	shmkeystatement {
	    pmsg ("service shmkey [%d]\n", $1.set[0].v.ival);
	    cur_service.shmkey = $1.set[0].v.ival;
	    free ($1.set);
	}
|
	connectiontimeoutstatement {
	    pmsg ("connection timout [%d]\n", $1.set[0].v.ival);
	    cur_service.connectiontimeout = $1.set[0].v.ival;
	    free ($1.set);
	}
|
	maxconnectionsstatement {
	    pmsg ("max clients in service [%d]\n", $1.set[0].v.ival);
	    cur_service.maxconnections = $1.set[0].v.ival;
	    free ($1.set);
	}
|
	typestatement {
	    pmsg ("service type [%d]\n", $1.set[0].v.ival);
	    cur_service.type = $1.set[0].v.ival;
	    free ($1.set);
	}
|
	inspectionstatement {
	    pmsg ("service header inspection [%d]\n", $1.set[0].v.ival);
	    cur_service.inspection = $1.set[0].v.ival;
	    free ($1.set);
	}
|
	allowfromstatement {
	    pmsg ("allow from [%s]\n", $1.set[0].v.sval);
	    add_allowfrom ($1.set[0].v.sval);
	    free ($1.set);
	}
|
	allowfilestatement {
	    pmsg ("allow file [%s]\n", $1.set[0].v.sval);
	    cur_service.allowfile = $1.set[0].v.sval;
	    free ($1.set);
	}
|
	denyfromstatement {
	    pmsg ("deny from [%s]\n", $1.set[0].v.sval);
	    add_denyfrom ($1.set[0].v.sval);
	    free ($1.set);
	}
|
	denyfilestatement {
	    pmsg ("deny file [%s]\n", $1.set[0].v.sval);
	    cur_service.denyfile = $1.set[0].v.sval;
	    free ($1.set);
	}
|
	useraccountstatement {
	    pmsg ("user account [%s]\n", $1.set[0].v.sval);
	    setuseraccount ($1.set[0].v.sval);
	    free ($1.set[0].v.sval);
	    free ($1.set);
	}
|
	backendblock {
	    pmsg ("converting backend statements, count is [%d]\n", $1.n);
	    for (i = 0; i < $1.n; i++)
		switch ($1.set[i].cf) {
		case cf_portspec:
		    pmsg ("backend port [%d]\n", $1.set[i].v.ival);
		    cur_backend.initial_port = $1.set[i].v.ival;
		    break;
		case cf_serverspec:
		    pmsg ("backend server [%s]\n", $1.set[i].v.sval);
		    cur_backend.initial_server = serverpart ($1.set[i].v.sval);
		    cur_backend.initial_port   = portpart ($1.set[i].v.sval);
		    free ($1.set[i].v.sval);
		    break;
		case cf_verbosityspec:
		    pmsg ("backend verbosity [%d]\n", $1.set[i].v.ival);
		    cur_backend.verbosity = $1.set[i].v.ival;
		    break;
		case cf_onstartspec:
		    pmsg ("backend onstart [%s]\n", $1.set[i].v.sval);
		    cur_backend.onstart = $1.set[i].v.sval;
		    break;
		case cf_onfailspec:
		    pmsg ("backend onfail [%s]\n", $1.set[i].v.sval);
		    cur_backend.onfail = $1.set[i].v.sval;
		    break;
		case cf_onendspec:
		    pmsg ("backend onend [%s]\n", $1.set[i].v.sval);
		    cur_backend.onend = $1.set[i].v.sval;
		    break;
		case cf_dumpspec:
		    pmsg ("backend trafficlog [%s]\n", $1.set[i].v.sval);
		    cur_backend.dumpfile = $1.set[i].v.sval;
		    break;
		case cf_thruspec:
		    pmsg ("backend throughputlog [%s]\n", $1.set[i].v.sval);
		    cur_backend.thruputfile = $1.set[i].v.sval;
		    break;
		case cf_httptimingspec:
		    if (cur_service.type != type_http)
			error ("'httptiminglog' only allowed in 'type http' "
			       "service definitions");
		    pmsg ("backend http timing log [%s]\n", $1.set[i].v.sval);
		    cur_backend.timinglog = $1.set[i].v.sval;
		    break;
		case cf_weightspec:
		    pmsg ("backend weight [%d]\n", $1.set[i].v.ival);
		    cur_backend.weight = $1.set[i].v.ival;
		    break;
		case cf_decayspec:
		    pmsg ("backend decay [%d]\n", $1.set[i].v.ival);
		    if ($1.set[i].v.ival >= 100)
			error ("Decay specifier %d must be a percentage, "
			       "never more than 99",
			       $1.set[i].v.ival);
		    cur_backend.decay = $1.set[i].v.ival;
		    break;
		case cf_maxconnectionsspec:
		    pmsg ("backend max clients [%d]\n", $1.set[i].v.ival);
		    cur_backend.maxconnections = $1.set[i].v.ival;
		    break;
		case cf_stickycookiespec:
		    pmsg ("backend sticky cookie [%s]\n",
			  $1.set[i].v.sval);
		    cur_backend.stickycookie = $1.set[i].v.sval;
		    break;
		case cf_addclientheaderspec:
		    pmsg ("client header to add [%s]\n",
			  $1.set[i].v.sval);
		    cur_backend.addclientheader =
			xrealloc (cur_backend.addclientheader,
				  (cur_backend.naddclientheader + 1) *
				  sizeof(char*));
		    cur_backend.addclientheader
			[cur_backend.naddclientheader++] =
			    $1.set[i].v.sval;
		    break;
		case cf_setclientheaderspec:
		    pmsg ("client header to set [%s]\n",
			  $1.set[i].v.sval);
		    cur_backend.setclientheader =
			xrealloc (cur_backend.setclientheader,
				  (cur_backend.nsetclientheader + 1) *
				  sizeof(char*));
		    cur_backend.setclientheader
			[cur_backend.nsetclientheader++] =
			    $1.set[i].v.sval;
		    break;
		case cf_appendclientheaderspec:
		    pmsg ("client header to append [%s]\n",
			  $1.set[i].v.sval);
		    cur_backend.appendclientheader =
			xrealloc (cur_backend.appendclientheader,
				  (cur_backend.nappendclientheader + 1) *
				  sizeof(char*));
		    cur_backend.appendclientheader
			[cur_backend.nappendclientheader++] =
			    $1.set[i].v.sval;
		    break;
		case cf_addserverheaderspec:
		    pmsg ("server header to add [%s]\n",
			  $1.set[i].v.sval);
		    cur_backend.addserverheader =
			xrealloc (cur_backend.addserverheader,
				  (cur_backend.naddserverheader + 1) *
				  sizeof(char*));
		    cur_backend.addserverheader
			[cur_backend.naddserverheader++] =
			    $1.set[i].v.sval;
		    break;
		case cf_setserverheaderspec:
		    pmsg ("server header to set [%s]\n",
			  $1.set[i].v.sval);
		    cur_backend.setserverheader =
			xrealloc (cur_backend.setserverheader,
				  (cur_backend.nsetserverheader + 1) *
				  sizeof(char*));
		    cur_backend.setserverheader
			[cur_backend.nsetserverheader++] =
			    $1.set[i].v.sval;
		    break;
		case cf_appendserverheaderspec:
		    pmsg ("server header to append:",
			  $1.set[i].v.sval);
		    cur_backend.appendserverheader =
			xrealloc (cur_backend.appendserverheader,
				  (cur_backend.nappendserverheader + 1) *
				  sizeof(char*));
		    cur_backend.appendserverheader
			[cur_backend.nappendserverheader++] =
			    $1.set[i].v.sval;
		    break;
		case cf_retriesspec:
		    pmsg ("backend retries [%d]\n", $1.set[i].v.ival);
		    cur_backend.retries = $1.set[i].v.ival;
		    break;
		case cf_statespec:
		    pmsg ("backend initial state [%d]\n", $1.set[i].v.ival);
		    cur_backend.initial_avail = $1.set[i].v.ival;
		    break;
		default:
		    error ("Internal jam, unhandled type %d "
			   "in backend specification",
			   $1.set[i].cf);
		}
	    free ($1.set);

	    /* Verify the backend block, supply defaults,
	     * And so on.
	     */
	    if (!cur_backend.initial_port)
		error ("Service %s: back end %s lacks port",
		       cur_service.name, cur_backend.name);
	    if (cur_service.dispatchext && cur_backend.maxconnections)
		error ("Service %s: back end %s defines maxconnections, "
		       "but an external dispatcher is in effect",
		       cur_service.name, cur_backend.name);
	    if (cur_backend.weight < 1)
		cur_backend.weight = 1;
	    if (cur_backend.retries < 1)
		cur_backend.retries = 1;
	    
	    /* Add to the list. */
	    cur_service.backend = xrealloc (cur_service.backend,
					    ++cur_service.nbackend *
					    sizeof(Backend));
	    cur_service.backend[cur_service.nbackend - 1] =
		cur_backend;
	    pmsg ("this was backend definition [%d]\n", cur_service.nbackend);
	    memset (&cur_backend, 0, sizeof(cur_backend));
	}
;

portstatement:
	PORT
	number
	semicol {
	    pmsg ("port statement [%d]\n", lastnr);
	    $$.n = 1;
	    $$.set = xmalloc (sizeof(Confset));
	    $$.set[0].cf = cf_portspec;
	    $$.set[0].v.ival = lastnr;
	}
;

bindstatement:
	BINDTO
	ipaddress
	semicol {
	    pmsg ("bindto statement [%d]\n", laststr);
	    $$.n = 1;
	    $$.set = xmalloc (sizeof(Confset));
	    $$.set[0].cf = cf_bindspec;
	    $$.set[0].v.sval = xstrdup(laststr);
	}
;

ipaddress:
	ipaddress_expected
	STRING {
	    setlaststr (laststring);
	    free (laststring);
	    laststring = 0;
	}
;

number:
	number_expected
	NUMBER {
	    setlastnr (SYMBOL);
	}
;

semicol:
	semicol_expected
	';'
;

verbositystatement:
	VERBOSITY
	onoff_expected
	onoff
	semicol {
	    pmsg ("verbosity statement [%d]\n", lastnr);
	    $$.n = 1;
	    $$.set = xmalloc (sizeof(Confset));
	    $$.set[0].cf = cf_verbosityspec;
	    $$.set[0].v.ival = lastnr;
	}
;

onoff:
	ON {
	    lastnr = 1;
	}
|
	OFF {
	    lastnr = 0;
	}
;

dispatchmodestatement:
	DISPATCHMODE
	dispatchmethod
	dispatchtail
	semicol {
	    $$ = $2;
	}
;

dispatchtail:
	dispatchover
	|
	dispatchext
	|
	/* empty */ {
	    $$.n = 0;
	}
;

dispatchover:
	OVER
	overnumber {
	    pmsg ("dispatch mode statement [%d]\n", lastnr);
	    $$.n = 1;
	    $$.set = xmalloc (sizeof(Confset));
	    $$.set[0].cf = cf_dispatchspec;
	    $$.set[0].v.ival = lastnr;

	    if (lastovernr &&
		(lastnr != ds_bysize && lastnr != ds_byduration))
		error ("Service '%s': this dispatch mode "
		       "doesn't support 'over <connections>'",
		       cur_service.name);
	}
;

overnumber:
	number_expected
	NUMBER {
	    setlastovernr (SYMBOL);
	}
;

dispatchext:
	commandline {
	    pmsg ("external handler [%s]\n", laststr);
	    if (lastnr != ds_externalhandler)
		error ("Service %s: this dispatch mode doesn't support "
		       "an external handler", cur_service.name);
	    setlastext (laststr);
	}
;

dispatchmethod:
	dispatchmethod_expected
	dispatchmethodspec {
	    $$.n = 1;
	    $$.set = xmalloc (sizeof(Confset));
	    $$.set[0].v.ival = lastnr;
	}
;

dispatchmethodspec:
	ROUNDROBIN {
	    lastnr = ds_roundrobin;
	}
|
	RANDOM {
	    lastnr = ds_random;
	}
|
	BYDURATION {
	    lastnr = ds_byduration;
	}
|
	BYSIZE {
	    lastnr = ds_bysize;
	}
|
	BYORDER {
	    lastnr = ds_byorder;
	}
|
	BYCONNECTIONS {
	    lastnr = ds_byconnections;
	}
|
	EXTERNALHANDLER {
	    lastnr = ds_externalhandler;
	}
|
	BYCLIENTIP {
	    lastnr = ds_byclientip;
	}
;

useraccountstatement:
	USERACCOUNT
	useraccount
	semicol {
	    pmsg ("user account statement [%d]\n", laststr);
	    $$.n = 1;
	    $$.set = xmalloc (sizeof(Confset));
	    $$.set[0].cf = cf_useraccountspec;
	    $$.set[0].v.sval = xstrdup (laststr);
	}
;

useraccount:
	useraccount_expected
	STRING {
	    setlaststr (laststring);
	    free (laststring);
	    laststring = 0;
	}
;

revivingintervalstatement:
	REVIVINGINTERVAL
	number
	opt_externalhandler
	semicol {
	    pmsg ("reviving interval statement [%d]\n", lastnr);
	    $$.n = 2;
	    $$.set = xmalloc (2 * sizeof(Confset));
	    $$.set[0].cf = cf_revivespec;
	    $$.set[0].v.ival = lastnr;
	    if ($3.n) {
		pmsg ("reviving externalhandler [%s]\n", $3.set[0].v.sval);
		$$.set[1].v.sval = $3.set[0].v.sval;
	    } else
		$$.set[1].v.sval = 0;
	}
;

checkintervalstatement:
	CHECKINTERVAL
	number
	opt_externalhandler
	semicol {
	    pmsg ("check interval [%d]\n", lastnr);
	    $$.n = 2;
	    $$.set = xmalloc (2 * sizeof(Confset));
	    $$.set[0].cf = cf_checkspec;
	    $$.set[0].v.ival = lastnr;
	    if ($3.n) {
		pmsg ("check externalhandler [%s]\n", $3.set[0].v.sval);
		$$.set[1].v.sval = $3.set[0].v.sval;
	    } else
		$$.set[1].v.sval = 0;
	}
;

opt_externalhandler:
	externalhandler
	|
	/* empty */ {
	    $$.n = 0;
	}
;

externalhandler:
	EXTERNALHANDLER
	commandline {
	    $$.n = 1;
	    $$.set = xmalloc (sizeof(Confset));
	    $$.set[0].v.sval = xstrdup (laststr);
	}
;

backlogstatement:
	BACKLOG
	number
	semicol {
	    pmsg ("backlog statement [%d]\n", lastnr);
	    $$.n = 1;
	    $$.set = xmalloc (sizeof(Confset));
	    $$.set[0].cf = cf_backlogspec;
	    $$.set[0].v.ival = lastnr;
	}
;

shmkeystatement:
	SHMKEY
	number
	semicol {
	    pmsg ("shmkey statement [%d]\n", lastnr);
	    $$.n = 1;
	    $$.set = xmalloc (sizeof(Confset));
	    $$.set[0].cf = cf_shmkeyspec;
	    $$.set[0].v.ival = lastnr;
	}
;

connectiontimeoutstatement:
	CONNECTIONTIMEOUT
	number
	semicol {
	    pmsg ("connection timeout statement [%d]\n", lastnr);
	    $$.n = 1;
	    $$.set = xmalloc (sizeof(Confset));
	    $$.set[0].cf = cf_connectiontimeoutspec;
	    $$.set[0].v.ival = lastnr;
	}
;

maxconnectionsstatement:
	MAXCONNECTIONS
	number
	semicol {
	    pmsg ("max clients statement (service) [%d]\n", lastnr);
	    $$.n = 1;
	    $$.set = xmalloc (sizeof(Confset));
	    $$.set[0].cf = cf_maxconnectionsspec;
	    $$.set[0].v.ival = lastnr;
	}
;

typestatement:
	TYPE
	typespec
	semicol {
	    pmsg ("service type [%d]\n", lastnr);
	    $$.n = 1;
	    $$.set = xmalloc (sizeof(Confset));
	    $$.set[0].cf = cf_typespec;
	    $$.set[0].v.ival = lastnr;
	}
;

typespec:
	type_expected
	typespecifier
	;

typespecifier:
	ANY {
	    lastnr = type_any;
	}
|
	HTTP {
	    lastnr = type_http;
	}
;

inspectionstatement:
	HEADERINSPECTION
	shallowdeepspec
	semicol {
	    pmsg ("service header inspection [%d]\n", lastnr);
	    $$.n = 1;
	    $$.set = xmalloc (sizeof(Confset));
	    $$.set[0].cf = cf_inspectionspec;
	    $$.set[0].v.ival = lastnr;
	}
;

shallowdeepspec:
	shallowdeep_expected
	shallowdeepspecifier
	;

shallowdeepspecifier:
	DEEP {
	    lastnr = ins_deep;
	}
|
	SHALLOW {
	    lastnr = ins_shallow;
	}
;

allowfromstatement:
	ALLOWFROM
	ipfilters
	semicol {
	    pmsg ("allow from [%s]\n", laststr);
	    $$.n = 1;
	    $$.set = xmalloc (sizeof(Confset));
	    $$.set[0].cf = cf_allowfromspec;
	    $$.set[0].v.sval = xstrdup(laststr);
	}
;

denyfromstatement:
	DENYFROM
	ipfilters
	semicol {
	    pmsg ("deny from [%s]\n", laststr);
	    $$.n = 1;
	    $$.set = xmalloc (sizeof(Confset));
	    $$.set[0].cf = cf_denyfromspec;
	    $$.set[0].v.sval = xstrdup(laststr);
	}
;

allowfilestatement:
	ALLOWFILE
	filename
	semicol {
	    pmsg ("allow file [%s]\n", laststr);
	    $$.n = 1;
	    $$.set = xmalloc (sizeof(Confset));
	    $$.set[0].cf = cf_allowfilespec;
	    $$.set[0].v.sval = xstrdup(laststr);
	}
;

denyfilestatement:
	DENYFILE
	filename
	semicol {
	    pmsg ("deny file [%s]\n", laststr);
	    $$.n = 1;
	    $$.set = xmalloc (sizeof(Confset));
	    $$.set[0].cf = cf_allowfilespec;
	    $$.set[0].v.sval = xstrdup(laststr);
	}
;

ipfilters:
	ipfilters_expected
	STRING {
	    setlaststr (laststring);
	    free (laststring);
	    laststring = 0;
	}
;

backendblock:
	BACKEND
	backendname
	openbrace
	backenddefinitions
	closebrace {
	    $$ = $4;
	}
;

backendname:
	backendname_expected
	IDENTIFIER {
	    pmsg ("backend name [%s]\n", SYMBOL);
	    for (i = 0; i < cur_service.nbackend; i++)
		if (!strcmp (cur_service.backend[i].name, SYMBOL))
		    error ("Service %s: back end '%s' multiply defined",
			   cur_service.name, SYMBOL);
	    cur_backend.name = xstrdup (SYMBOL);
	}
;

backenddefinitions:
	backenddefinitions
	backenddefinition {
	    $1.n++;
	    $1.set = xrealloc ($1.set, $1.n * sizeof(Confset));
	    $1.set[$1.n - 1] = $2.set[0];
	    $$ = $1;
	}
|
	backenddefinition {
	    $$ = $1;
	}
;

backenddefinition:
	backendstatement_expected
	backendstatement {
	    $$ = $2;
	}
;

backendstatement:
	serverstatement {
	    pmsg ("backend server [%s]\n", $1.set[0].v.sval);
	    $$ = $1;
	}
|
	portstatement {
	    pmsg ("backend port [%d]\n", $1.set[0].v.ival);
	    $$ = $1;
	}
|
	verbositystatement {
	    pmsg ("backend verbosity [%d]\n", $1.set[0].v.ival);
	    $$ = $1;
	}
|
	onstartstatement {
	    pmsg ("backend onstart [%s]\n", $1.set[0].v.sval);
	    $$ = $1;
	}
|
	onendstatement {
	    pmsg ("backend onend [%s]\n", $1.set[0].v.sval);
	    $$ = $1;
	}
|
	onfailstatement {
	    pmsg ("backend onfail [%s]\n", $1.set[0].v.sval);
	    $$ = $1;
	}
|
	dumptrafficstatement {
	    pmsg ("backend trafficlog [%s]\n", $1.set[0].v.sval);
	    $$ = $1;
	}
|
	throughputstatement {
	    pmsg ("backend trafficlog [%s]\n", $1.set[0].v.sval);
	    $$ = $1;
	}
|
	httptiminglogstatement {
	    pmsg ("backend HTTP timing log [%s]\n", $1.set[0].v.sval);
	    $$ = $1;
	}
|
	weightstatement {
	    pmsg ("backend weight [%d]\n", $1.set[0].v.ival);
	    $$ = $1;
	}
|
	decaystatement {
	    pmsg ("backend decay [%d]\n", $1.set[0].v.ival);
	    $$ = $1;
	}
|
	maxconnectionsstatement {
	    pmsg ("backend maxconnections [%d]\n", $1.set[0].v.ival);
	    $$ = $1;
	}
|
	stickycookiestatement {
	    pmsg ("backend sticky cookie [%s]\n", $1.set[0].v.sval);
	    $$ = $1;
	}
|
	addclientheaderstatement {
	    pmsg ("addclientheader [%s]\n", $1.set[0].v.sval);
	    $$ = $1;
	}
|
	setclientheaderstatement {
	    pmsg ("setclientheader [%s]\n", $1.set[0].v.sval);
	    $$ = $1;
	}
|
	appendclientheaderstatement {
	    pmsg ("appendclientheader [%s]\n", $1.set[0].v.sval);
	    $$ = $1;
	}
|
	addserverheaderstatement {
	    pmsg ("addserverheader [%s]\n", $1.set[0].v.sval);
	    $$ = $1;
	}
|
	setserverheaderstatement {
	    pmsg ("setserverheader [%s]\n", $1.set[0].v.sval);
	    $$ = $1;
	}
|
	appendserverheaderstatement {
	    pmsg ("appendserverheader [%s]\n", $1.set[0].v.sval);
	    $$ = $1;
	}
|
	retriesstatement {
	    pmsg ("backend retries [%d]\n", $1.set[0].v.ival);
	    $$ = $1;
	}
|
	initialstatestatement {
	    pmsg ("backend state [%d]\n", $1.set[0].v.ival);
	    $$ = $1;
	}
;

serverstatement:
	SERVER
	serveraddress_expected
	serveraddress
	semicol {
	    pmsg ("server statement [%s]\n", laststr);
	    $$.n = 1;
	    $$.set = xmalloc (sizeof (Confset));
	    $$.set[0].cf = cf_serverspec;
	    $$.set[0].v.sval = xstrdup (laststr);
	}
;

weightstatement:
	WEIGHT
	number
	semicol {
	    pmsg ("weight statement [%d]\n", lastnr);
	    $$.n = 1;
	    $$.set = xmalloc (sizeof (Confset));
	    $$.set[0].cf = cf_weightspec;
	    $$.set[0].v.ival = lastnr;
	}
;

decaystatement:
	DECAY
	number
	semicol {
	    pmsg ("decay statement [%d]\n", lastnr);
	    $$.n = 1;
	    $$.set = xmalloc (sizeof (Confset));
	    $$.set[0].cf = cf_decayspec;
	    $$.set[0].v.ival = lastnr;
	}
;

serveraddress:
	STRING {
	    setlaststr (laststring);
	}
;

retriesstatement:
	RETRIES
	number
	semicol {
	    pmsg ("retries [%d]\n", lastnr);
	    $$.n = 1;
	    $$.set = xmalloc (sizeof(Confset));
	    $$.set[0].cf = cf_retriesspec;
	    $$.set[0].v.ival = lastnr;
	}
;

initialstatestatement:
	STATE
	statedef
	semicol {
	    pmsg ("state [%d]\n", lastnr);
	    $$.n = 1;
	    $$.set = xmalloc (sizeof(Confset));
	    $$.set[0].cf = cf_statespec;
	    $$.set[0].v.ival = lastnr;
	}
;

statedef:
	statedef_expected
	somestate
;

somestate:
	AVAILABLE {
	    lastnr = st_available;
	}
|
	UNAVAILABLE {
	    lastnr = st_unavailable;
	}
|
	DOWN {
	    lastnr = st_down;
	}
;

onstartstatement:
	ONSTART
	commandline
	semicol {
	    pmsg ("onstart statement [%s]\n", laststr);
	    $$.n = 1;
	    $$.set = xmalloc (sizeof (Confset));
	    $$.set[0].cf = cf_onstartspec;
	    $$.set[0].v.sval = xstrdup (laststr);
	}
;

onfailstatement:
	ONFAIL
	commandline
	semicol {
	    pmsg ("onfail statement [%s]\n", laststr);
	    $$.n = 1;
	    $$.set = xmalloc (sizeof (Confset));
	    $$.set[0].cf = cf_onfailspec;
	    $$.set[0].v.sval = xstrdup (laststr);
	}
;

onendstatement:
	ONEND
	commandline
	semicol {
	    pmsg ("onend statement [%s]\n", laststr);
	    $$.n = 1;
	    $$.set = xmalloc (sizeof (Confset));
	    $$.set[0].cf = cf_onendspec;
	    $$.set[0].v.sval = xstrdup (laststr);
	}
;

dumptrafficstatement:
	TRAFFICLOG
	filename
	semicol {
	    pmsg ("trafficlog statement [%s]\n", laststr);
	    $$.n = 1;
	    $$.set = xmalloc (sizeof (Confset));
	    $$.set[0].cf = cf_dumpspec;
	    $$.set[0].v.sval = xstrdup (laststr);
	}
;

throughputstatement:
	THROUGHPUTLOG
	filename
	semicol {
	    pmsg ("throughputlog statement [%s]\n", laststr);
	    $$.n = 1;
	    $$.set = xmalloc (sizeof (Confset));
	    $$.set[0].cf = cf_thruspec;
	    $$.set[0].v.sval = xstrdup (laststr);
	}
;

httptiminglogstatement:
	HTTPTIMINGLOG
	filename
	semicol {
	    pmsg ("httptiminglog statement [%s]\n", laststr);
	    $$.n = 1;
	    $$.set = xmalloc (sizeof (Confset));
	    $$.set[0].cf = cf_httptimingspec;
	    $$.set[0].v.sval = xstrdup (laststr);
	}
;

commandline:
	commandline_expected
	STRING {
	    setlaststr (laststring);
	    free (laststring);
	    laststring = 0;
	}
;

filename:
	filename_expected
	STRING {
	    setlaststr (laststring);
	    free (laststring);
	    laststring = 0;
	}
;

stickycookiestatement:
	STICKYCOOKIE
	cookiespecifier
	semicol {
	    pmsg ("insertcookie statement [%s]\n", laststr);
	    $$.n = 1;
	    $$.set = xmalloc (sizeof (Confset));
	    $$.set[0].cf = cf_stickycookiespec;
	    $$.set[0].v.sval = xstrdup (laststr);
	}
;

cookiespecifier:
	cookie_expected
	STRING {
		setlaststr (laststring);
		free (laststring);
		laststring = 0;
	}
;

addclientheaderstatement:
	ADDCLIENTHEADER
	headerstring
	semicol {
	    pmsg ("addclientheader statement [%s]\n", laststr);
	    $$.n = 1;
	    $$.set = xmalloc (sizeof(Confset));
	    $$.set[0].cf = cf_addclientheaderspec;
	    $$.set[0].v.sval = xstrdup (laststr);
	}
;

setclientheaderstatement:
	SETCLIENTHEADER
	headerstring
	semicol {
	    pmsg ("setclientheader statement [%s]\n", laststr);
	    $$.n = 1;
	    $$.set = xmalloc (sizeof(Confset));
	    $$.set[0].cf = cf_setclientheaderspec;
	    $$.set[0].v.sval = xstrdup (laststr);
	}
;

appendclientheaderstatement:
	APPENDCLIENTHEADER
	headerstring
	semicol {
	    pmsg ("appendclientheader statement [%s]\n", laststr);
	    $$.n = 1;
	    $$.set = xmalloc (sizeof(Confset));
	    $$.set[0].cf = cf_appendclientheaderspec;
	    $$.set[0].v.sval = xstrdup (laststr);
	}
;

addserverheaderstatement:
	ADDSERVERHEADER
	headerstring
	semicol {
	    pmsg ("addserverheader statement [%s]\n", laststr);
	    $$.n = 1;
	    $$.set = xmalloc (sizeof(Confset));
	    $$.set[0].cf = cf_addserverheaderspec;
	    $$.set[0].v.sval = xstrdup (laststr);
	}
;

setserverheaderstatement:
	SETSERVERHEADER
	headerstring
	semicol {
	    pmsg ("setserverheader statement [%s]\n", laststr);
	    $$.n = 1;
	    $$.set = xmalloc (sizeof(Confset));
	    $$.set[0].cf = cf_setserverheaderspec;
	    $$.set[0].v.sval = xstrdup (laststr);
	}
;

appendserverheaderstatement:
	APPENDSERVERHEADER
	headerstring
	semicol {
	    pmsg ("appendserverheader statement [%s]\n", laststr);
	    $$.n = 1;
	    $$.set = xmalloc (sizeof(Confset));
	    $$.set[0].cf = cf_appendserverheaderspec;
	    $$.set[0].v.sval = xstrdup (laststr);
	}
;

headerstring:
	headerstring_expected
	STRING {
	    setlaststr (laststring);
	    free (laststring);
	    laststring = 0;
	}
;

headerstring_expected: {
    yyerrmsg = "HTTP header specifier expected";
}
;

cookie_expected: {
    yyerrmsg = "cookie specifier expected";
}
;

number_expected: {
    yyerrmsg = "number expected";
}
;

serveraddress_expected: {
    yyerrmsg = "hostname or IP address expected";
}
;

service_expected: {
    yyerrmsg = "'service' expected";
}
;

backendstatement_expected: {
    yyerrmsg = "backend definition statement expected";
}
;

servicebody_expected: {
    yyerrmsg = "service body statement expected";
}
;

semicol_expected: {
    yyerrmsg = "semicolon (;) expected";
}
;

onoff_expected: {
    yyerrmsg = "'on' or 'off' expetcted";
}
;

dispatchmethod_expected: {
    yyerrmsg = "dispatch method expected";
}
;

commandline_expected: {
    yyerrmsg = "command line expected";
}
;

filename_expected: {
    yyerrmsg = "file name expected";
}
;

servicename_expected: {
    yyerrmsg = "service name (identifier) expected";
}
;

backendname_expected: {
    yyerrmsg = "backend name (identifier) expected";
}
;

ipaddress_expected: {
    yyerrmsg = "IP address such as 1.2.3.4 or 'any' expected";
}
;

type_expected: {
    yyerrmsg = "Service type expected ('any', 'stickyhttp', ...)";
}
;

shallowdeep_expected: {
    yyerrmsg = "Header inspection mode mode expected ('shallow' or 'deep')";
}
;

ipfilters_expected: {
    yyerrmsg = "IP filter(s) expected";
}
;

useraccount_expected: {
    yyerrmsg = "username expected";
}
;

statedef_expected: {
    yyerrmsg = "state definition expected";
}
;

openbrace:
	openbrace_expected
	'{'
;

closebrace:
	closebrace_expected
	'}'
;

openbrace_expected: {
    yyerrmsg = "'{' expected";
}

closebrace_expected: {
    yyerrmsg = "'}' expected";
}    

option_statement_expected: {
    yyerrmsg = "option statement (logactivity, tcpbuffersize, ...) expected";
}    
