/*************************************************************************
 * This file is part of Crosroads 1.79, a load balancer and fail over
 * utility for TCP. Copyright (c) Karel Kubat, distributed under GPL.
 * Visit http://crossroads.e-tunity.com for information.
 *************************************************************************/

#include "../crossroads.h"

char *fmt_expand (char const *s, int *skip, int fmt_number) {
    static char buf[80];
    int nextnumber;

    /* Expansion table:
     * %a	- availability of current back end (if any), 0 not available,
     *		  %1a = first back end etc.
     * %b	- name of current back end (if any), %1b = first back end etc.
     * %e	- seconds since epoch, number is an addition count 
     * %g       - GMT format timestamp. The time in between is the minutes
     *		  offset into the future, so %30g is now + 1/2 hour
     * %h	- actual serveraddress of current back end, %h1 = first etc.
     * %l	- last connect time of current back end, %l1 = first etc.
     * %n	- nr of clients of current back end, %1n = first back end etc.
     * %p 	- actual port of current back end, %1p = first etc.
     * %r 	- client IP, number is not interpreted
     * %s	- name of service, number is not interpreted
     * %t, %T	- ANSI timestamp of local or GMT time, number is offset in min
     *		  into the future
     * %v	- Crossroads version, number not interpreted
     * %w	- weight of current back end, %1w = first back end etc.
     */

    (*skip)++;
    /* 
    msg ("Fmt expand [%s], skip=%d, nr=%d",
	 s, *skip, fmt_number);
    */
    
    switch (*s) {

    case 'a':
	if (!fmt_number)
	    fmt_number = current_backend;
	else
	    fmt_number--;
	if (fmt_number < 0 || fmt_number >= activeservice->nbackend ||
	    !backend_available (fmt_number))
	    return ("0");
	return ("1");
	
    case 'b':
	if (!fmt_number)
	    fmt_number = current_backend;
	else
	    fmt_number--;
	if (fmt_number >= 0 && fmt_number < activeservice->nbackend)
	    return (activeservice->backend[fmt_number].name);
	return ("noname");

    case 'e':
	snprintf (buf, sizeof(buf) - 1, "%u", (unsigned) time(0) + fmt_number);
	return (buf);

    case 'g':
	return (gmtstamp (fmt_number));

    case 'h':
	if (!fmt_number)
	    fmt_number = current_backend;
	else
	    fmt_number--;
	if (fmt_number < 0 || fmt_number >= activeservice->nbackend)
	    return ("noname");
	return (servicereport->backendstate[fmt_number].actual_server);

    case 'l':
	if (!fmt_number)
	    fmt_number = current_backend;
	else
	    fmt_number--;
	if (fmt_number >= 0 && fmt_number < activeservice->nbackend) {
	    snprintf (buf, sizeof(buf) - 1, "%f",
		      servicereport->backendstate[fmt_number].last_con_sec);
	    return (buf);
	}
	return ("0");

    case 'n':
	if (!fmt_number)
	    fmt_number = current_backend;
	else
	    fmt_number--;
	if (fmt_number >= 0 && fmt_number < activeservice->nbackend) {
	    snprintf (buf, sizeof(buf) - 1, "%u",
		      (unsigned)
		        servicereport->backendstate[fmt_number].nclients);
	    return (buf);
	}
	return ("0");

    case 'p':
	if (!fmt_number)
	    fmt_number = current_backend;
	else
	    fmt_number--;
	if (fmt_number < 0 || fmt_number >= activeservice->nbackend)
	    return ("0");
	snprintf (buf, sizeof(buf), "%d",
		  servicereport->backendstate[fmt_number].actual_port);
	return (buf);
	
    case 'r':
	if (client_ip && *client_ip)
	    return (client_ip);
	return ("0.0.0.0");

    case 's':
	return (activeservice->name);

    case 't':
	return (ansistamp (tm_localtime, fmt_number));

    case 'T':
	return (ansistamp (tm_gmtime, fmt_number));

    case 'v':
	return (VER);

    case 'w':
	if (!fmt_number)
	    fmt_number = current_backend;
	else
	    fmt_number--;
	if (fmt_number >= 0 && fmt_number < activeservice->nbackend) {
	    snprintf (buf, sizeof(buf) - 1, "%u",
		      (unsigned)
		        activeservice->backend[fmt_number].weight);
	    return (buf);
	}
	return ("0");

    case '0':
    case '1':
    case '2':
    case '3':
    case '4':
    case '5':
    case '6':
    case '7':
    case '8':
    case '9':
	nextnumber = atoi(s);
	while (isdigit (*s)) {
	    (*skip)++;
	    s++;
	}
	(*skip)--;
	return (fmt_expand (s, skip, nextnumber));
	
    default:
	*buf = *(s + 1);
	*(buf + 1) = 0;
	return (buf);
    }
}

char *str_expand_format (char const *h) {
    char *ret = 0;
    char const *cp;
    int skip;

    for (cp = h; cp && *cp; cp++) {
	if (*cp == '%') {
	    skip = 0;
	    ret = xstrcat (ret, fmt_expand (cp + 1, &skip, 0));
	    cp += skip;
	} else {
	    ret = xstrcatch (ret, *cp);
	}
    }

    /*
    msg ("String [%s] expanded to [%s]",
	 h, ret);
    */
    return (ret);
}
